import os
import numpy as np
import shutil
import torch 
from time import time
import networkx as nx

from pdb2sql import interface

from .tools import BioWrappers, PSSM, BSA
from .Graph import Graph


class ResidueGraph(Graph):

    def __init__(self, pdb=None, pssm=None,
                 contact_distance=8.5, internal_contact_distance=3,
                 pssm_align='res', biopython=False):
        """Class from which Residue features are computed

        Args:
            pdb (str, optional): path to pdb file. Defaults to None.
            pssm (str, optional): path to pssm file. Defaults to None.
            contact_distance (float, optional): cutoff distance for external edges. Defaults to 8.5.
            internal_contact_distance (int, optional): cutoff distance for internal edges. Defaults to 3.
            pssm_align (str, optional): [description]. Defaults to 'res'.
        """
        super().__init__()

        self.type = 'residue'
        self.pdb = pdb
        self.name = os.path.splitext(os.path.basename(pdb))[0]

        if pssm is not None:
            self.pssm, self.ic = PSSM.PSSM_aligned(
                pssm, style=pssm_align)
        else:
            self.pssm, self.ic = None, None

        self.contact_distance = contact_distance
        self.internal_contact_distance = internal_contact_distance

        self.residue_charge = {'CYS': -0.64, 'HIS': -0.29, 'ASN': -1.22, 'GLN': -1.22, 'SER': -0.80, 'THR': -0.80, 'TYR': -0.80,
                               'TRP': -0.79, 'ALA': -0.37, 'PHE': -0.37, 'GLY': -0.37, 'ILE': -0.37, 'VAL': -0.37, 'MET': -0.37,
                               'PRO': 0.0, 'LEU': -0.37, 'GLU': -1.37, 'ASP': -1.37, 'LYS': -0.36, 'ARG': -1.65}

        self.residue_names = {'CYS': 0, 'HIS': 1, 'ASN': 2, 'GLN': 3, 'SER': 4, 'THR': 5, 'TYR': 6, 'TRP': 7,
                              'ALA': 8, 'PHE': 9, 'GLY': 10, 'ILE': 11, 'VAL': 12, 'MET': 13, 'PRO': 14, 'LEU': 15,
                              'GLU': 16, 'ASP': 17, 'LYS': 18, 'ARG': 19}

        self.residue_polarity = {'CYS': 'polar', 'HIS': 'polar', 'ASN': 'polar', 'GLN': 'polar', 'SER': 'polar', 'THR': 'polar', 'TYR': 'polar', 'TRP': 'polar',
                                 'ALA': 'apolar', 'PHE': 'apolar', 'GLY': 'apolar', 'ILE': 'apolar', 'VAL': 'apolar', 'MET': 'apolar', 'PRO': 'apolar', 'LEU': 'apolar',
                                 'GLU': 'neg_charged', 'ASP': 'neg_charged', 'LYS': 'neg_charged', 'ARG': 'pos_charged'}

        self.pssm_pos = {'CYS': 4, 'HIS': 8, 'ASN': 2, 'GLN': 5, 'SER': 15, 'THR': 16, 'TYR': 18, 'TRP': 17,
                         'ALA': 0, 'PHE': 13, 'GLY': 7, 'ILE': 9, 'VAL': 19, 'MET': 12, 'PRO': 14, 'LEU': 10,
                         'GLU': 6, 'ASP': 3, 'LYS': 11, 'ARG': 1}

        self.polarity_encoding = {
            'apolar': 0, 'polar': 1, 'neg_charged': 2, 'pos_charged': 3}
        
        self.biopython = biopython
        #self.edge_polarity_encoding, iencod = {}, 0
        ##for k1, v1 in self.polarity_encoding.items():
        ##for k2, v2 in self.polarity_encoding.items():
        ##key = tuple(np.sort([v1, v2]))
        ##if key not in self.edge_polarity_encoding:
        ##self.edge_polarity_encoding[key] = iencod
        #iencod += 1

        # check if external execs are installed
        self.check_execs()

        # create the sqldb
        db = interface(self.pdb)

        # get the graphs
        t0 = time()
        self.get_graph(db)
        print('Graph %f' % (time()-t0))

        # get the nodes/edge attributes
        t0 = time()
        self.get_node_features(db)
        print('Node %f' % (time()-t0))

        t0 = time()
        self.get_edge_features()
        print('Edge %f' % (time()-t0))

        # close the db
        db._close()

    def check_execs(self):
        """Checks if all the required external execs are installed.

        Raises:
            OSError: if execs not found
        """

        execs = {'msms': 'http://mgltools.scripps.edu/downloads#msms'}

        for e, inst in execs.items():
            if shutil.which(e) is None:
                raise OSError(e, ' is not installed see ',
                              inst, ' for details')

    def get_graph(self, db):
        """Gets the interface graph nodes and edges given the 
        internal and external distance threshold

        Args:
            db ([type]): sqldb database
        """
        self.nx = nx.Graph()
        self.nx.edge_index = []
        self.res_contact_pairs = db.get_contact_residues(
            cutoff=self.contact_distance, return_contact_pairs=True)

        # get all the nodes
        all_nodes = self._get_all_valid_nodes(
            self.res_contact_pairs, self.pssm)

        # create the interface edges
        for key, val in self.res_contact_pairs.items():
            if key in all_nodes:
                for v in val:
                    if v in all_nodes:
                        d = self._get_edge_distance(key, v, db)
                        self.nx.add_edge(key, v, dist=d, type=bytes(
                            'interface', encoding='utf-8'))
                    else:
                        print('WARNING: {} is not a valid node'.format(v))
            else:
                print('WARNING: {} is not a valid node'.format(key))

        # get the internal edges
        edges, dist = self.get_internal_edges(db)
        for e, d in zip(edges, dist):
            e0, e1 = tuple(e[0]), tuple(e[1])
            if e0 in all_nodes and e1 in all_nodes:
                self.nx.add_edge(e0, e1, dist=d, type=bytes(
                    'internal', encoding='utf-8'))
            else:
                print('WARNING: {} {} is not a valid edge'.format(e0, e1))

    @staticmethod
    def _get_all_valid_nodes(res_contact_pairs, pssm, verbose=False):
        """Filters out invalid nodes based on 2 criteria:
            - contains unrecognized residue(s)
            - contains residue(s) absent form the pssm matrix

        Args:
            res_contact_pairs ([type]): list of contact pairs 
            pssm ([type]): pssm file
            verbose (bool, optional): [description]. Defaults to False.

        Returns:
            Valid list of contact pairs
        """
        valid_res = [
            'ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLU', 'GLN', 'GLY', 'HIS', 'ILE', 'LEU', 'LYS', 'MET', 'PHE',
            'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'ASX', 'SEC', 'GLX']

        # tag the non residues
        keys_to_pop = []
        for res in res_contact_pairs.keys():
            if res[2] not in valid_res:
                keys_to_pop.append(res)
                # res_contact_pairs.pop(res,None)
                Warning('--> Residue ', res, ' not valid')

        # tag the ones that are not in PSSM
        if pssm is not None:
            for res in list(res_contact_pairs.keys()):
                if res not in pssm:
                    keys_to_pop.append(res)
                    # res_contact_pairs.pop(res,None)
                    Warning('--> Residue ', res,
                            ' not found in PSSM file')

        # Remove the residue
        for res in keys_to_pop:
            if res in res_contact_pairs:
                res_contact_pairs.pop(res, None)

        # get a list of residues of chain B
        # automatically remove the ones that are not proper residues
        # and the ones that are not in the PSSM
        nodesB = []
        for k, reslist in list(res_contact_pairs.items()):

            for res in reslist:
                if res[2] in valid_res:
                    if pssm is None:
                        nodesB += [res]
                    elif res in pssm:
                        nodesB += [res]
                else:
                    if verbose:
                        print('removed node', res)
        nodesB = sorted(set(nodesB))

        # make a list of nodes
        return list(res_contact_pairs.keys()) + nodesB

    def get_node_features(self, db):
        """Assigns features to each node
        """

        # get the BSA of the residues
        t0 = time()
        bsa_calc = BSA.BSA(self.pdb, db)
        bsa_calc.get_structure()
        bsa_calc.get_contact_residue_sasa(
            cutoff=self.contact_distance)
        bsa_data = bsa_calc.bsa_data
        #print('_BSA %f' %(time()-t0))

        # biopython data
        t0 = time()
        model = BioWrappers.get_bio_model(db.pdbfile)
        #print('_Model %f' %(time()-t0))
        
        if self.biopython == True :
            #t0 = time()
            ResDepth = BioWrappers.get_depth_contact_res(
                model, self.nx.nodes)
            #print('_RD %f' %(time()-t0))
            
            #t0 = time()
            HSE = BioWrappers.get_hse(model)
            #print('_HSE %f' %(time()-t0))
            
        # loop over all the nodes
        for node_key in self.nx.nodes:

            chainID = node_key[0]
            resName = node_key[2]
            resSeq = int(node_key[1])

            self.nx.nodes[node_key]['chain'] = {
                'A': 0, 'B': 1}[chainID]
            self.nx.nodes[node_key]['pos'] = np.mean(
                db.get('x,y,z', chainID=chainID, resSeq=resSeq), 0)
            self.nx.nodes[node_key]['type'] = self.onehot(
                self.residue_names[resName], len(self.residue_names))

            self.nx.nodes[node_key]['charge'] = self.residue_charge[resName]
            self.nx.nodes[node_key]['polarity'] = self.onehot(
                self.polarity_encoding[self.residue_polarity[resName]], len(self.polarity_encoding))

            self.nx.nodes[node_key]['bsa'] = bsa_data[node_key]

            if self.pssm is not None:
                data = PSSM.get_pssm_data(node_key, self.pssm)
                self.nx.nodes[node_key]['pssm'] = data
                self.nx.nodes[node_key]['cons'] = data[self.pssm_pos[resName]]
                self.nx.nodes[node_key]['ic'] = PSSM.get_ic_data(
                    node_key, self.ic)
            
            if self.biopython == True :

                self.nx.nodes[node_key]['depth'] = ResDepth[node_key] if node_key in ResDepth else 0
                bio_key = (chainID, resSeq)
                self.nx.nodes[node_key]['hse'] = HSE[bio_key] if bio_key in HSE else (
                    0, 0, 0)

    def get_edge_features(self):
        """Assigns distance feature to each edge
        """
        node_keys = list(self.nx.nodes)

        for e in self.nx.edges:
            node1, node2 = e
            #self.nx.edges[node1, node2]['polarity'] = self._get_edge_polarity(
            #    node1, node2)
            self.nx.edge_index.append(
                [node_keys.index(node1), node_keys.index(node2)])

    def get_internal_edges(self, db):
        """Gets internal edges and the associated distances 
        """
        nodesA, nodesB = [], []
        for n in self.nx.nodes:
            if n[0] == 'A':
                nodesA.append(n)
            elif n[0] == 'B':
                nodesB.append(n)

        edgesA, distA = self._get_internal_edges_chain(
            nodesA, db, self.internal_contact_distance)
        edgesB, distB = self._get_internal_edges_chain(
            nodesB, db, self.internal_contact_distance)

        return edgesA+edgesB, distA+distB

    def _get_internal_edges_chain(self, nodes, db, cutoff):
        """Gets internal edges and the associated distances (min) per chain

        Args:
            nodes ([type]): list of nodes
            db ([type]): sqldb database
            cutoff ([type]): internal edge cutoff

        Returns:
            edges between two nodes, 
            and the minimum distance between those two nodes
        """
        lnodes = list(nodes)
        nn = len(nodes)
        edges, dist = [], []
        for i1 in range(nn):
            xyz1 = np.array(
                db.get('x,y,z', chainID=nodes[i1][0], resSeq=nodes[i1][1]))
            for i2 in range(i1+1, nn):
                xyz2 = np.array(
                    db.get('x,y,z', chainID=nodes[i2][0], resSeq=nodes[i2][1]))
                d2 = -2*np.dot(xyz1, xyz2.T) + np.sum(xyz1**2,
                                                      axis=1)[:, None] + np.sum(xyz2**2, axis=1)
                if np.any(d2 < cutoff**2):
                    edges.append((nodes[i1], nodes[i2]))
                    dist.append(np.sqrt(np.min(d2)))

        return edges, dist

    def _get_internal_edges_chain_mean(self, nodes, db, cutoff):
        """Gets internal edges and the associated distance (mean) per chain

        Args:
            nodes ([type]): list of nodes
            db ([type]): sqldb database
            cutoff ([type]): internal edge cutoff

        Returns:
            edges between two nodes, 
            and the mean distance between those two nodes
        """
        lnodes = list(nodes)
        nn = len(nodes)
        edges, dist, centers = [], [], []

        for i1 in range(nn):
            centers.append(
                np.mean(db.get('x,y,z', chainID=nodes[i1][0], resSeq=nodes[i1][1]), 0))

        centers = np.array(centers)
        distances = -2*np.dot(centers, centers.T) + np.sum(
            centers**2, axis=1)[:, None] + np.sum(centers**2, axis=1)

        indexes = np.argwhere(distances < 2*cutoff)

        for i, j in indexes:
            if i != j:
                edges.append((nodes[i], nodes[j]))
                dist.append(distances[i, j])

        return edges, dist

    def _get_edge_polarity(self, node1, node2):
        """Gets edge polarity

        Args:
            node1 
            node2 
        """
        v1 = self.nx.nodes[node1]['polarity']
        v2 = self.nx.nodes[node2]['polarity']

        key = tuple(np.sort([v1, v2]))
        return self.edge_polarity_encoding[key]

    def _get_edge_distance(self, node1, node2, db):

        # pos1 = self.nx.nodes[node1]['pos']
        # pos2 = self.nx.nodes[node2]['pos']
        # return np.linalg.norm(pos1-pos2)
        xyz1 = np.array(
            db.get('x,y,z', chainID=node1[0], resSeq=node1[1]))
        xyz2 = np.array(
            db.get('x,y,z', chainID=node2[0], resSeq=node2[1]))
        d2 = -2*np.dot(xyz1, xyz2.T) + np.sum(xyz1**2,
                                              axis=1)[:, None] + np.sum(xyz2**2, axis=1)
        return np.sqrt(np.min(d2))


    def onehot(self, idx, size):
        """One hot encoder
        """
        onehot = torch.zeros(size)
        # Fill the one-hot encoded sequence with 1 at the corresponding idx
        onehot[idx] = 1
        return np.array(onehot)



"""Module containing some utilities for testing."""

from pathlib import Path
import pkg_resources as pkg

__all__ = ["PATH_DEEPRANK_GNN", "PATH_TEST"]

# Environment data
PATH_DEEPRANK_GNN = Path(pkg.resource_filename('deeprank_gnn', ''))
ROOT = PATH_DEEPRANK_GNN.parent

PATH_TEST = ROOT / "tests"



import os
from pdb2sql import StructureSimilarity
import numpy as np
import networkx as nx
from collections import OrderedDict
import time
from deeprank_gnn.tools.embedding import manifold_embedding
import community
import markov_clustering as mc
import h5py


class Graph(object):

    def __init__(self):
        """Class that performs graph level action

            - get score for the graph (lrmsd, irmsd, fnat, capri_class, bin_class and dockQ)
            - networkx object (graph) to hdf5 format
            - networkx object (graph) from hdf5 format
        """
        self.name = None
        self.nx = None
        self.score = {'irmsd': None, 'lrmsd': None, 'capri_class': None,
                      'fnat': None, 'dockQ': None, 'bin_class': None}

    def get_score(self, ref):
        """Assigns scores (lrmsd, irmsd, fnat, dockQ, bin_class, capri_class) to a protein graph

        Args:
            ref (path): path to the reference structure required to compute the different score
        """

        ref_name = os.path.splitext(os.path.basename(ref))[0]
        sim = StructureSimilarity(self.pdb, ref)
        
        # Input pre-computed zone files
        if os.path.exists(ref_name+'.lzone'):
            self.score['lrmsd'] = sim.compute_lrmsd_fast(
            method='svd', lzone=ref_name+'.lzone')
            self.score['irmsd'] = sim.compute_irmsd_fast(
            method='svd', izone=ref_name+'.izone')
        
        # Compute zone files
        else: 
            self.score['lrmsd'] = sim.compute_lrmsd_fast(
            method='svd')
            self.score['irmsd'] = sim.compute_irmsd_fast(
            method='svd')

        self.score['fnat'] = sim.compute_fnat_fast()
        self.score['dockQ'] = sim.compute_DockQScore(
            self.score['fnat'], self.score['lrmsd'], self.score['irmsd'])
        self.score['bin_class'] = self.score['irmsd'] < 4.0
        
        self.score['capri_class'] = 5
        for thr, val in zip([6.0, 4.0, 2.0, 1.0], [4, 3, 2, 1]):
            if self.score['irmsd'] < thr:
                self.score['capri_class'] = val
        
    def nx2h5(self, f5):
        """Converts Networkx object to hdf5 format

        Args:
            f5 ([type]): hdf5 file
        """
        # group for the graph
        grp = f5.create_group(self.name)

        # store the nodes
        data = np.array(list(self.nx.nodes)).astype('S')
        grp.create_dataset('nodes', data=data)

        # store node features
        node_feat_grp = grp.create_group('node_data')
        feature_names = list(list(self.nx.nodes.data())[0][1].keys())
        for feat in feature_names:
            data = [v for _, v in nx.get_node_attributes(
                self.nx, feat).items()]
            node_feat_grp.create_dataset(feat, data=data)

        edges, internal_edges = [], []
        edge_index, internal_edge_index = [], []

        edge_feat = list(list(self.nx.edges.data())[0][2].keys())
        edge_data, internal_edge_data = {}, {}
        for feat in edge_feat:
            edge_data[feat] = []
            internal_edge_data[feat] = []

        node_key = list(self.nx.nodes)

        for e in self.nx.edges:

            edge_type = self.nx.edges[e]['type'].decode('utf-8')
            ind1, ind2 = node_key.index(e[0]), node_key.index(e[1])

            if edge_type == 'interface':
                edges.append(e)
                edge_index.append([ind1, ind2])

                for feat in edge_feat:
                    edge_data[feat].append(self.nx.edges[e][feat])

            elif edge_type == 'internal':
                internal_edges.append(e)
                internal_edge_index.append([ind1, ind2])

                for feat in edge_feat:
                    internal_edge_data[feat].append(
                        self.nx.edges[e][feat])

        # store the edges
        data = np.array(edges).astype('S')
        grp.create_dataset('edges', data=data)

        data = np.array(internal_edges).astype('S')
        grp.create_dataset('internal_edges', data=data)

        # store the edge index
        grp.create_dataset('edge_index', data=edge_index)
        grp.create_dataset('internal_edge_index',
                           data=internal_edge_index)

        # store the edge attributes
        edge_feat_grp = grp.create_group('edge_data')
        internal_edge_feat_grp = grp.create_group(
            'internal_edge_data')

        for feat in edge_feat:
            edge_feat_grp.create_dataset(feat, data=edge_data[feat])
            internal_edge_feat_grp.create_dataset(
                feat, data=internal_edge_data[feat])

        # store the score
        score_grp = grp.create_group('score')
        for k, v in self.score.items():
            if v is not None:
                score_grp.create_dataset(k, data=v)

    def h52nx(self, f5name, mol, molgrp=None):
        """Converts Hdf5 file to Networkx object

        Args:
            f5name (str): hdf5 file
            mol (str): molecule name
            molgrp ([type], optional): hdf5[molecule]. Defaults to None.
        """
        if molgrp is None:

            f5 = h5py.File(f5name, 'r')
            molgrp = f5[mol]
            self.name = mol
            self.pdb = mol+'.pdb'

        else:
            self.name = molgrp.name
            self.pdb = self.name+'.pdb'

        # creates the graph
        self.nx = nx.Graph()

        # get nodes
        nodes = molgrp['nodes'][()].astype('U').tolist()
        nodes = [tuple(n) for n in nodes]

        # get node features
        node_keys = list(molgrp['node_data'].keys())
        node_feat = {}
        for key in node_keys:
            node_feat[key] = molgrp['node_data/'+key][()]

        # add nodes
        for iN, n in enumerate(nodes):
            self.nx.add_node(n)
            for k, v in node_feat.items():
                v = v.reshape(-1, 1) if v.ndim == 1 else v
                v = v[iN, :]
                v = v[0] if len(v) == 1 else v
                self.nx.nodes[n][k] = v

        # get edges
        edges = molgrp['edges'][()].astype('U').tolist()
        edges = [(tuple(e[0]), tuple(e[1])) for e in edges]

        # get edge data
        edge_key = list(molgrp['edge_data'].keys())
        edge_feat = {}
        for key in edge_key:
            edge_feat[key] = molgrp['edge_data/'+key][()]

        # add edges
        for iedge, e in enumerate(edges):
            self.nx.add_edge(e[0], e[1])
            for k, v in edge_feat.items():
                v = v.reshape(-1, 1) if v.ndim == 1 else v
                v = v[iedge, :]
                v = v[0] if len(v) == 1 else v
                self.nx.edges[e[0], e[1]][k] = v

        # get internal edges
        edges = molgrp['internal_edges'][()].astype(
            'U').tolist()
        edges = [(tuple(e[0]), tuple(e[1])) for e in edges]

        # get edge data
        edge_key = list(molgrp['internal_edge_data'].keys())
        edge_feat = {}
        for key in edge_key:
            edge_feat[key] = molgrp['internal_edge_data/' +
                                    key][()]

        # add edges
        for iedge, e in enumerate(edges):
            self.nx.add_edge(e[0], e[1])
            for k, v in edge_feat.items():
                v = v.reshape(-1, 1) if v.ndim == 1 else v
                v = v[iedge, :]
                v = v[0] if len(v) == 1 else v
                self.nx.edges[e[0], e[1]][k] = v

        # add score
        self.score = {}
        score_key = list(molgrp['score'].keys())
        for key in score_key:
            self.score[key] = molgrp['score/'+key][()]

        # add cluster
        if 'clustering' in molgrp:
            self.clusters = {}
            for method in list(molgrp['clustering'].keys()):
                self.clusters[method] = molgrp['clustering/' +
                                               method+'/depth_0'][()]

        if molgrp is None:
            f5.close()

    def plotly_2d(self, out=None, offline=False, iplot=True, method='louvain'):
        """Plots the interface graph in 2D

        Args:
            out ([type], optional): output name. Defaults to None.
            offline (bool, optional): Defaults to False.
            iplot (bool, optional): Defaults to True.
            method (str, optional): 'mcl' of 'louvain'. Defaults to 'louvain'.
        """
        if offline:
            import plotly.offline as py
        else:
            import chart_studio.plotly as py

        import plotly.graph_objs as go
        import matplotlib.pyplot as plt

        pos = np.array(
            [v.tolist() for _, v in nx.get_node_attributes(self.nx, 'pos').items()])
        pos2D = manifold_embedding(pos)
        dict_pos = {n: p for n, p in zip(self.nx.nodes, pos2D)}
        nx.set_node_attributes(self.nx, dict_pos, 'pos2D')

        # remove interface edges for clustering
        gtmp = self.nx.copy()
        ebunch = []
        for e in self.nx.edges:
            typ = self.nx.edges[e]['type']
            if isinstance(typ, bytes):
                typ = typ.decode('utf-8')
            if typ == 'interface':
                ebunch.append(e)
        gtmp.remove_edges_from(ebunch)

        try:

            raw_cluster = self.clusters[method]
            cluster = {}
            node_key = list(self.nx.nodes.keys())

            for inode, index_cluster in enumerate(raw_cluster):
                cluster[node_key[inode]] = index_cluster
        except:

            print(
                'No cluster in the group. Computing cluster now with method : %s' % method)

            if method == 'louvain':
                cluster = community.best_partition(gtmp)

            elif method == 'mcl':
                matrix = nx.to_scipy_sparse_matrix(gtmp)
                # run MCL with default parameters
                result = mc.run_mcl(matrix)
                mcl_clust = mc.get_clusters(result)    # get clusters
                cluster = {}
                node_key = list(self.nx.nodes.keys())
                for ic, c in enumerate(mcl_clust):
                    for node in c:
                        cluster[node_key[node]] = ic

        # get the colormap for the clsuter line
        ncluster = np.max([v for _, v in cluster.items()])+1
        cmap = plt.cm.nipy_spectral
        N = cmap.N
        cmap = [cmap(i) for i in range(N)]
        cmap = cmap[::int(N/ncluster)]
        cmap = 'plasma'

        edge_trace_list, internal_edge_trace_list = [], []
        node_connect = {}

        for edge in self.nx.edges:

            edge_type = self.nx.edges[edge[0],
                                      edge[1]]['type'].decode('utf-8')
            if edge_type == 'internal':
                trace = go.Scatter(x=[], y=[], text=[], mode='lines', hoverinfo=None,  showlegend=False,
                                   line=go.scatter.Line(color='rgb(110,110,110)', width=3))
            elif edge_type == 'interface':
                trace = go.Scatter(x=[], y=[], text=[], mode='lines', hoverinfo=None,  showlegend=False,
                                   line=go.scatter.Line(color='rgb(210,210,210)', width=1))

            x0, y0 = self.nx.nodes[edge[0]]['pos2D']
            x1, y1 = self.nx.nodes[edge[1]]['pos2D']

            trace['x'] += (x0, x1, None)
            trace['y'] += (y0, y1, None)

            if edge_type == 'internal':
                internal_edge_trace_list.append(trace)
            elif edge_type == 'interface':
                edge_trace_list.append(trace)

            for i in [0, 1]:
                if edge[i] not in node_connect:
                    node_connect[edge[i]] = 1
                else:
                    node_connect[edge[i]] += 1
        node_trace_A = go.Scatter(x=[], y=[], text=[], mode='markers', hoverinfo='text',
                                  marker=dict(color='rgb(227,28,28)', size=[],
                                              line=dict(color=[], width=4, colorscale=cmap)))
        # 'rgb(227,28,28)'
        node_trace_B = go.Scatter(x=[], y=[], text=[], mode='markers', hoverinfo='text',
                                  marker=dict(color='rgb(0,102,255)', size=[],
                                              line=dict(color=[], width=4, colorscale=cmap)))
        # 'rgb(0,102,255)'
        node_trace = [node_trace_A, node_trace_B]

        for node in self.nx.nodes:

            index = self.nx.nodes[node]['chain']
            pos = self.nx.nodes[node]['pos2D']

            node_trace[index]['x'] += (pos[0],)
            node_trace[index]['y'] += (pos[1],)
            node_trace[index]['text'] += (
                '[Clst:' + str(cluster[node]) + '] ' + ' '.join(node),)

            nc = node_connect[node]
            node_trace[index]['marker']['size'] += (
                5 + 15*np.tanh(nc/5),)
            node_trace[index]['marker']['line']['color'] += (
                cluster[node],)

        fig = go.Figure(data=[*internal_edge_trace_list, *edge_trace_list, *node_trace],
                        layout=go.Layout(
            title='<br>tSNE connection graph for %s' % self.pdb,
            titlefont=dict(size=16),
            showlegend=False,
            hovermode='closest',
            margin=dict(b=20, l=5, r=5, t=40),
            annotations=[dict(
                text="",
                showarrow=False,
                xref="paper", yref="paper",
                x=0.005, y=-0.002)],
            xaxis=dict(
                showgrid=False, zeroline=False, showticklabels=False),
            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False)))

        if iplot:
            py.iplot(fig, filename=out)
        else:
            py.plot(fig)

    def plotly_3d(self, out=None, offline=False, iplot=True):
        """Plots interface graph in 3D

        Args:
            out ([type], optional): [description]. Defaults to None.
            offline (bool, optional): [description]. Defaults to False.
            iplot (bool, optional): [description]. Defaults to True.
        """
        if offline:
            import plotly.offline as py
        else:
            import chart_studio.plotly as py

        import plotly.graph_objs as go

        edge_trace_list, internal_edge_trace_list = [], []
        node_connect = {}

        for edge in self.nx.edges:

            edge_type = self.nx.edges[edge[0],
                                      edge[1]]['type'].decode('utf-8')
            if edge_type == 'internal':
                trace = go.Scatter3d(x=[], y=[], z=[], text=[], mode='lines', hoverinfo=None,  showlegend=False,
                                     line=go.scatter3d.Line(color='rgb(110,110,110)', width=5))
            elif edge_type == 'interface':
                trace = go.Scatter3d(x=[], y=[], z=[], text=[], mode='lines', hoverinfo=None,  showlegend=False,
                                     line=go.scatter3d.Line(color='rgb(210,210,210)', width=2))

            x0, y0, z0 = self.nx.nodes[edge[0]]['pos']
            x1, y1, z1 = self.nx.nodes[edge[1]]['pos']

            trace['x'] += (x0, x1, None)
            trace['y'] += (y0, y1, None)
            trace['z'] += (z0, z1, None)

            if edge_type == 'internal':
                internal_edge_trace_list.append(trace)
            elif edge_type == 'interface':
                edge_trace_list.append(trace)

            for i in [0, 1]:
                if edge[i] not in node_connect:
                    node_connect[edge[i]] = 1
                else:
                    node_connect[edge[i]] += 1
        node_trace_A = go.Scatter3d(x=[], y=[], z=[], text=[], mode='markers', hoverinfo='text',
                                    marker=dict(color='rgb(227,28,28)', size=[], symbol='circle',
                                                line=dict(color='rgb(50,50,50)', width=2)))

        node_trace_B = go.Scatter3d(x=[], y=[], z=[], text=[], mode='markers', hoverinfo='text',
                                    marker=dict(color='rgb(0,102,255)', size=[], symbol='circle',
                                                line=dict(color='rgb(50,50,50)', width=2)))

        node_trace = [node_trace_A, node_trace_B]

        for node in self.nx.nodes:

            index = self.nx.nodes[node]['chain']
            pos = self.nx.nodes[node]['pos']

            node_trace[index]['x'] += (pos[0],)
            node_trace[index]['y'] += (pos[1],)
            node_trace[index]['z'] += (pos[2], )
            node_trace[index]['text'] += (' '.join(node),)

            nc = node_connect[node]
            node_trace[index]['marker']['size'] += (
                5 + 15*np.tanh(nc/5), )

        fig = go.Figure(data=[*node_trace, *internal_edge_trace_list, *edge_trace_list],
                        layout=go.Layout(
                        title='<br>Connection graph for %s' % self.pdb,
                        titlefont=dict(size=16),
                        showlegend=False,
                        hovermode='closest',
                        margin=dict(b=20, l=5, r=5, t=40),
                        annotations=[dict(
                            text="",
                            showarrow=False,
                            xref="paper", yref="paper",
                            x=0.005, y=-0.002)],
                        xaxis=dict(
                            showgrid=False, zeroline=False, showticklabels=False),
                        yaxis=dict(showgrid=False, zeroline=False, showticklabels=False)))

        if iplot:
            py.iplot(fig, filename=out)
        else:
            py.plot(fig)


if __name__ == "__main__":
    import h5py
    graph = Graph()
    graph.h52nx('1AK4_residue.hdf5', '1ATN')
    graph.plotly_2d('1ATN')



import os
import sys
import h5py
from tqdm import tqdm
import time

from .ResidueGraph import ResidueGraph
from .Graph import Graph


class GraphHDF5(object):

    def __init__(self, pdb_path, ref_path=None, graph_type='residue', pssm_path=None,
                 select=None, outfile='graph.hdf5', biopython=False):
        """Master class from which graphs are computed 

        Args:
            pdb_path (str): path to the docking models
            ref_path (str, optional): path to the reference model. Defaults to None.
            graph_type (str, optional): Defaults to 'residue'.
            pssm_path ([type], optional): path to the pssm file. Defaults to None.
            select (str, optional): filter files that starts with 'input'. Defaults to None.
            outfile (str, optional): Defaults to 'graph.hdf5'.
        """
        # get the list of PDB names
        pdbs = list(filter(lambda x: x.endswith(
            '.pdb'), os.listdir(pdb_path)))
        if select is not None:
            pdbs = list(filter(lambda x: x.startswith(select), pdbs))

        f5 = h5py.File(outfile, 'w')

        desc = '{:25s}'.format('   Create HDF5')
        data_tqdm = tqdm(pdbs, desc=desc, file=sys.stdout)

        for name in data_tqdm:

            # pdb name
            pdbfile = os.path.join(pdb_path, name)

            # mol name and base name
            mol_name = os.path.splitext(name)[0]
            base_name = mol_name.split('_')[0]

            if graph_type == 'residue':
                # get the pssm file for pdb
                if pssm_path is not None: 
                    pssm = self._get_pssm(pssm_path, mol_name, base_name)
                else pssm = None
                # generate a graph 
                graph = ResidueGraph(pdb=pdbfile, pssm=pssm, biopython=biopython)

            # get the score
            if ref_path is not None:
                ref = os.path.join(ref_path, base_name+'.pdb')
                graph.get_score(ref)

            # export the graph in hdf5 format
            try:
                graph.nx2h5(f5)
            except:
                print('WARNING: No graph generated for {}'.format(name))

        f5.close()

    @staticmethod
    def _get_pssm(pssm_path, mol_name, base_name):
        """Gets the pssm file for a given model

        Args:
            pssm_path (str): path to the pssm file
            mol_name (str): molecule name (XXXX_A.pdb -> XXXX_A)
            base_name (str): short molecule name (XXXX_A.pdb -> XXXX)

        Returns:
            pssm files for chain A and for chain B 
            pssm = {'A': pssmA, 'B': pssmB}
        """
        pssmA = os.path.join(pssm_path, base_name+'.A.pdb.pssm')
        pssmB = os.path.join(pssm_path, base_name+'.B.pdb.pssm')

        # check if the pssms exists
        if os.path.isfile(pssmA) and os.path.isfile(pssmB):
            pssm = {'A': pssmA, 'B': pssmB}
        else:
            pssmA = os.path.join(pssm_path, base_name+'.A.pssm')
            pssmB = os.path.join(pssm_path, base_name+'.B.pssm')
            if os.path.isfile(pssmA) and os.path.isfile(pssmB):
                pssm = {'A': pssmA, 'B': pssmB}
            else:
                raise FileNotFoundError(
                    'PSSM file for ' + mol_name + ' not found')
        return pssm

if __name__ == '__main__':

    pdb_path = './data/pdb/1ATN/'
    pssm_path = './data/pssm/1ATN/'
    ref = './data/ref/1ATN/'

    GraphHDF5(pdb_path=pdb_path, ref_path=ref, pssm_path=pssm_path,
              graph_type='residue', outfile='1AK4_residue.hdf5')



import torch
from torch.nn import Parameter
import torch.nn.functional as F
import torch.nn as nn


class EGAT(torch.nn.Module):
    """
    Graph Attention Layer including edge features
    Inspired by Sazan Mahbub et al. "EGAT: Edge Aggregated Graph Attention Networks and Transfer Learning Improve Protein-Protein Interaction Site Prediction"
    """

    def __init__(self,
                 in_channels,
                 out_channels, 
                 number_edge_features=1,
                 bias=False):

        super(EGAT, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.fc = nn.Linear(self.in_channels, self.out_channels, bias=bias)
        self.fc_edge_attr = nn.Linear(number_edge_features, 3, bias=bias)
        self.fc_attention = nn.Linear(2 * self.out_channels + 3, 1, bias=bias)
        self.initialize_parameters()
        
    def initialize_parameters(self):
        size = self.in_channels
        uniform(size, self.fc.weight)
        uniform(size, self.fc_attention.weight)
        uniform(size, self.fc_edge_attr.weight)
        
    def forward(self, x, edge_index, edge_attr):

        row, col = edge_index
        num_node = len(x)
        edge_attr = edge_attr.unsqueeze(
            -1) if edge_attr.dim() == 1 else edge_attr

        xcol = self.fc(x[col])
        xrow = self.fc(x[row])
        
        ed = self.fc_edge_attr(edge_attr)
        # create edge feature by concatenating node feature
        alpha = torch.cat([xrow, xcol, ed], dim=1)
        alpha = self.fc_attention(alpha)
        alpha = F.leaky_relu(alpha)
        
        alpha = F.softmax(alpha, dim=1)
        h = alpha * xcol 
        
        out = torch.zeros(
            num_node, self.out_channels).to(h.device)
        z = scatter_sum(h, row, dim=0, out=out)
        
        return z
    
    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.in_channels,
                                   self.out_channels)



import os
import sys
import glob
import h5py
from tqdm import tqdm
import time
import multiprocessing as mp
from functools import partial
import pickle

from .ResidueGraph import ResidueGraph
from .Graph import Graph


class GraphHDF5(object):

    def __init__(self, pdb_path, ref_path=None, graph_type='residue', pssm_path=None,
                 select=None, outfile='graph.hdf5', nproc=1, use_tqdm=True, tmpdir='./',
                 limit=None, biopython=False):
        """Master class from which graphs are computed 
        Args:
            pdb_path (str): path to the docking models
            ref_path (str, optional): path to the reference model. Defaults to None.
            graph_type (str, optional): Defaults to 'residue'.
            pssm_path ([type], optional): path to the pssm file. Defaults to None.
            select (str, optional): filter files that starts with 'input'. Defaults to None.
            outfile (str, optional): Defaults to 'graph.hdf5'.
            nproc (int, optional): number of processors. Default to 1.
            use_tqdm (bool, optional): Default to True.
            tmpdir (str, optional): Default to `./`.
            limit (int, optional): Default to None.
        """
        # get the list of PDB names
        pdbs = list(filter(lambda x: x.endswith(
            '.pdb'), os.listdir(pdb_path)))
        if select is not None:
            pdbs = list(filter(lambda x: x.startswith(select), pdbs))

        # get the full path of the pdbs
        pdbs = [os.path.join(pdb_path, name) for name in pdbs]
        if limit is not None:
            if isinstance(limit, list):
                pdbs = pdbs[limit[0]:limit[1]]
            else:
                pdbs = pdbs[:limit]

        # get the pssm data
        pssm = {}
        for p in pdbs:
            base = os.path.basename(p)
            mol_name = os.path.splitext(base)[0]
            base_name = mol_name.split('_')[0]
            if pssm_path is not None:
                pssm[p] = self._get_pssm(pssm_path, mol_name, base_name)
            else:
                pssm[p] = None

        # get the ref path
        if ref_path is None:
            ref = None
        else:
            ref = os.path.join(ref_path, base_name+'.pdb')

        # compute all the graphs on 1 core and directly
        # store the graphs the HDF5 file
        if nproc == 1:
            graphs = self.get_all_graphs(
                pdbs, pssm, ref, outfile, use_tqdm, biopython)

        else:

            if not os.path.isdir(tmpdir):
                os.mkdir(tmpdir)

            pool = mp.Pool(nproc)
            part_process = partial(
                self._pickle_one_graph, pssm=pssm, ref=ref, tmpdir=tmpdir, biopython=biopython)
            pool.map(part_process, pdbs)

            # get teh graph names
            graph_names = [os.path.join(tmpdir, f)
                           for f in os.listdir(tmpdir)]
            graph_names = list(
                filter(lambda x: x.endswith('.pkl'), graph_names))
            if select is not None:
                graph_names = list(
                    filter(lambda x: x.startswith(tmpdir+select), graph_names))

            # transfer them to the hdf5
            f5 = h5py.File(outfile, 'w')
            desc = '{:25s}'.format('   Store in HDF5')

            for name in graph_names:
                f = open(name, 'rb')
                g = pickle.load(f)
                try:
                    g.nx2h5(f5)
                except Exception as e:
                    print('Issue encountered while computing graph ', name)
                    print(e)
                f.close()
                os.remove(name)
                
            f5.close()

        # clean up
        rmfiles = glob.glob(
            '*.izone') + glob.glob('*.lzone') + glob.glob('*.refpairs') 
        for f in rmfiles:
            os.remove(f)

    def get_all_graphs(self, pdbs, pssm, ref, outfile, use_tqdm=True, biopython=False):

        graphs = []
        if use_tqdm:
            desc = '{:25s}'.format('   Create HDF5')
            lst = tqdm(pdbs, desc=desc, file=sys.stdout)
        else:
            lst = pdbs

        for name in lst:
            try:
                graphs.append(self._get_one_graph(name, pssm, ref, biopython))
            except Exception as e:
                print('Issue encountered while computing graph ', name)
                print(e)

        f5 = h5py.File(outfile, 'w')
        for g in graphs:
            try:
                g.nx2h5(f5)
            except Exception as e:
                print('Issue encountered while storing graph ', g.pdb)
                print(e)
        f5.close()

    @staticmethod
    def _pickle_one_graph(name, pssm, ref, tmpdir='./', biopython=False):

        # get the graph
        g = ResidueGraph(pdb=name, pssm=pssm[name], biopython=biopython)
        if ref is not None:
            g.get_score(ref)

        # pickle it
        mol_name = os.path.basename(name)
        mol_name = os.path.splitext(mol_name)[0]
        fname = os.path.join(tmpdir, mol_name+'.pkl')

        f = open(fname, 'wb')
        pickle.dump(g, f)
        f.close()

    @staticmethod
    def _get_one_graph(name, pssm, ref, biopython):

        # get the graph
        g = ResidueGraph(pdb=name, pssm=pssm[name], biopython=biopython)
        if ref is not None:
            g.get_score(ref)
        return g

    @staticmethod
    def _get_pssm(pssm_path, mol_name, base_name):

        if pssm_path is None:
            return None

        pssmA = os.path.join(pssm_path, base_name+'.A.pssm')
        pssmB = os.path.join(pssm_path, base_name+'.B.pssm')

        # check if the pssms exists
        if os.path.isfile(pssmA) and os.path.isfile(pssmB):
            pssm = {'A': pssmA, 'B': pssmB}
        else:
            pssmA = os.path.join(pssm_path, base_name+'.A.pdb.pssm')
            pssmB = os.path.join(pssm_path, base_name+'.B.pdb.pssm')
            if os.path.isfile(pssmA) and os.path.isfile(pssmB):
                pssm = {'A': pssmA, 'B': pssmB}
            else:
                raise FileNotFoundError(
                    'PSSM file for ' + mol_name + ' not found')
        return pssm


if __name__ == '__main__':

    pdb_path = './data/pdb/1ATN/'
    pssm_path = './data/pssm/1ATN/'
    ref = './data/ref/1ATN/'

    GraphHDF5(pdb_path=pdb_path, ref_path=ref, pssm_path=pssm_path,
              graph_type='residue', outfile='1AK4_residue.hdf5')



import sys
import os
import torch
import numpy as np
from torch_geometric.data.dataset import Dataset
from torch_geometric.data.data import Data
from tqdm import tqdm
import h5py
import copy

from .community_pooling import community_detection, community_pooling


def DivideDataSet(dataset, percent=[0.8, 0.2], shuffle=True):
    """Divides the dataset into a training set and an evaluation set

    Args:
        dataset ([type])
        percent (list, optional): [description]. Defaults to [0.8, 0.2].
        shuffle (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """
    size = dataset.__len__()
    index = np.arange(size)

    if shuffle:
        np.random.shuffle(index)

    size1 = int(percent[0]*size)
    index1, index2 = index[:size1], index[size1:]

    dataset1 = copy.deepcopy(dataset)
    dataset1.index_complexes = [
        dataset.index_complexes[i] for i in index1]

    dataset2 = copy.deepcopy(dataset)
    dataset2.index_complexes = [
        dataset.index_complexes[i] for i in index2]

    return dataset1, dataset2


def PreCluster(dataset, method):
    """Pre-clusters nodes of the graphs

    Args:
        dataset (HDF5DataSet object)
        method (srt): 'mcl' (Markov Clustering) or 'louvain'
    """
    for fname, mol in tqdm(dataset.index_complexes):

        data = dataset.load_one_graph(fname, mol)

        if data is None:
            f5 = h5py.File(fname, 'a')
            try:
                print('deleting {}'.format(mol))
                del f5[mol]
            except:
                print('{} not found'.format(mol))
            f5.close()
            continue

        f5 = h5py.File(fname, 'a')
        grp = f5[mol]

        clust_grp = grp.require_group('clustering')

        if method.lower() in clust_grp:
            print('Deleting previous data for mol %s method %s' %
                  (mol, method))
            del clust_grp[method.lower()]

        method_grp = clust_grp.create_group(method.lower())

        cluster = community_detection(
            data.internal_edge_index, data.num_nodes, method=method)
        method_grp.create_dataset('depth_0', data=cluster)

        data = community_pooling(cluster, data)

        cluster = community_detection(
            data.internal_edge_index, data.num_nodes, method=method)
        method_grp.create_dataset('depth_1', data=cluster)

        f5.close()


class HDF5DataSet(Dataset):

    def __init__(self, root='./', database=None, transform=None, pre_transform=None,
                 dict_filter=None, target=None, tqdm=True, index=None,
                 node_feature='all', edge_feature=['dist'], clustering_method='mcl',
                 edge_feature_transform=lambda x: np.tanh(-x/2+2)+1):
        """Class from which the hdf5 datasets are loaded.

        Args:
            root (str, optional): [description]. Defaults to './'.
            database (str, optional): Path to hdf5 file(s). Defaults to None.
            transform (callable, optional): A function/transform that takes in an torch_geometric.data.Data object and returns a transformed version. The data object will be transformed before every access. Defaults to None.
            pre_transform (callable, optional):  A function/transform that takes in an torch_geometric.data.Data object and returns a transformed version. The data object will be transformed before being saved to disk.. Defaults to None.
            dict_filter dictionnary, optional): Dictionnary of type [name: cond] to filter the molecules. Defaults to None.
            target (str, optional): irmsd, lrmsd, fnat, bin, capri_class or DockQ. Defaults to None.
            tqdm (bool, optional): Show progress bar. Defaults to True.
            index (int, optional): index of a molecule. Defaults to None.
            node_feature (str or list, optional): consider all pre-computed node features ('all') or some defined node features (provide a list). Defaults to 'all'.
            edge_feature (list, optional): only distances are available in this version of DeepRank-GNN. Defaults to ['dist'].
            clustering_method (str, optional): 'mcl' (Markov Clustering) or 'louvain'. Defaults to 'mcl'.
            edge_feature_transform (function, optional): transformation applied to the edge features. Defaults to lambdax:np.tanh(-x/2+2)+1.
        """
        super().__init__(root, transform, pre_transform)

        # allow for multiple database
        self.database = database
        if not isinstance(database, list):
            self.database = [database]

        self.target = target
        self.dict_filter = dict_filter
        self.tqdm = tqdm
        self.index = index
        self.node_feature = node_feature
        self.edge_feature = edge_feature

        self.edge_feature_transform = edge_feature_transform

        self.clustering_method = clustering_method

        # check if the files are ok
        self.check_hdf5_files()

        # check the selection of features
        self.check_node_feature()
        self.check_edge_feature()

        # create the indexing system
        # alows to associate each mol to an index
        # and get fname and mol name from the index
        self.create_index_molecules()

    def len(self):
        """Gets the length of the dataset
        Returns:
            int: number of complexes in the dataset
        """
        return len(self.index_complexes)

    def _download(self):
        pass

    def _process(self):
        pass

    def get(self, index):
        """Gets one item from its unique index.

        Args:
            index (int): index of the complex
        Returns:
            dict: {'mol':[fname,mol],'feature':feature,'target':target}
        """

        fname, mol = self.index_complexes[index]
        data = self.load_one_graph(fname, mol)
        return data

    def check_hdf5_files(self):
        """Checks if the data contained in the hdf5 file is valid."""
        print("   Checking dataset Integrity")
        remove_file = []
        for fname in self.database:
            try:
                f = h5py.File(fname, 'r')
                mol_names = list(f.keys())
                if len(mol_names) == 0:
                    print('    -> %s is empty ' % fname)
                    remove_file.append(fname)
                f.close()
            except Exception as e:
                print(e)
                print('    -> %s is corrupted ' % fname)
                remove_file.append(fname)

        for name in remove_file:
            self.database.remove(name)

    def check_node_feature(self):
        """Checks if the required node features exist
        """
        f = h5py.File(self.database[0], 'r')
        mol_key = list(f.keys())[0]
        self.available_node_feature = list(
            f[mol_key+'/node_data/'].keys())
        f.close()

        if self.node_feature == 'all':
            self.node_feature = self.available_node_feature
        else:
            for feat in self.node_feature:
                if feat not in self.available_node_feature:
                    print(
                        feat, ' node feature not found in the file', self.database[0])
                    print('Possible node feature : ')
                    print('\n'.join(self.available_node_feature))
                    #raise ValueError('Feature Not found')
                    exit()

    def check_edge_feature(self):
        """Checks if the required edge features exist
        """
        f = h5py.File(self.database[0], 'r')
        mol_key = list(f.keys())[0]
        self.available_edge_feature = list(
            f[mol_key+'/edge_data/'].keys())
        f.close()

        if self.edge_feature == 'all':
            self.edge_feature = self.available_edge_feature
        elif self.edge_feature is not None:
            for feat in self.edge_feature:
                if feat not in self.available_edge_feature:
                    print(
                        feat, ' edge attribute not found in the file', self.database[0])
                    print('Possible edge attribute : ')
                    print('\n'.join(self.available_edge_feature))
                    #raise ValueError('Feature Not found')
                    exit()

    def load_one_graph(self, fname, mol):
        """Loads one graph

        Args:
            fname (str): hdf5 file name
            mol (str): name of the molecule

        Returns:
            Data object or None: torch_geometric Data object containing the node features, the internal and external edge features, the target and the xyz coordinates. Return None if features cannot be loaded.
        """
        f5 = h5py.File(fname, 'r')
        try:
            grp = f5[mol]
        except:
            f5.close()
            return None

        # nodes
        data = ()
        try:
            for feat in self.node_feature:
                vals = grp['node_data/'+feat][()]
                if vals.ndim == 1:
                    vals = vals.reshape(-1, 1)
                data += (vals,)
            x = torch.tensor(np.hstack(data), dtype=torch.float)

        except:
            print('node attributes not found in the file',
                  self.database[0])
            f5.close()
            return None

        try:
            # index ! we have to have all the edges i.e : (i,j) and (j,i)
            ind = grp['edge_index'][()]
            ind = np.vstack((ind, np.flip(ind, 1))).T
            edge_index = torch.tensor(
                ind, dtype=torch.long).contiguous()

            # edge feature (same issue than above)
            data = ()
            if self.edge_feature is not None:
                for feat in self.edge_feature:
                    vals = grp['edge_data/'+feat][()]
                    if vals.ndim == 1:
                        vals = vals.reshape(-1, 1)
                    data += (vals,)
                data = np.hstack(data)
                data = np.vstack((data, data))
                data = self.edge_feature_transform(data)
                edge_attr = torch.tensor(
                    data, dtype=torch.float).contiguous()

            else:
                edge_attr = None

            # internal edges
            ind = grp['internal_edge_index'][()]
            ind = np.vstack((ind, np.flip(ind, 1))).T
            internal_edge_index = torch.tensor(
                ind, dtype=torch.long).contiguous()

            # internal edge feature
            data = ()
            if self.edge_feature is not None:
                for feat in self.edge_feature:
                    vals = grp['internal_edge_data/'+feat][()]
                    if vals.ndim == 1:
                        vals = vals.reshape(-1, 1)
                    data += (vals,)
                data = np.hstack(data)
                data = np.vstack((data, data))
                data = self.edge_feature_transform(data)
                internal_edge_attr = torch.tensor(
                    data, dtype=torch.float).contiguous()

            else:
                internal_edge_attr = None

        except:
            print('edge features not found in the file',
                  self.database[0])
            f5.close()
            return None

        # target
        if self.target is None:
            y = None

        else:
            if grp['score/'+self.target][()] is not None:
                y = torch.tensor(
                    [grp['score/'+self.target][()]], dtype=torch.float).contiguous()
            else:
                y = None

        # pos
        pos = torch.tensor(grp['node_data/pos/']
                           [()], dtype=torch.float).contiguous()

        # load
        data = Data(x=x,
                    edge_index=edge_index,
                    edge_attr=edge_attr,
                    y=y,
                    pos=pos)

        data.internal_edge_index = internal_edge_index
        data.internal_edge_attr = internal_edge_attr

        # mol name
        data.mol = mol

        # cluster
        if 'clustering' in grp.keys():
            if self.clustering_method in grp['clustering'].keys():
                if ('depth_0' in grp['clustering/{}'.format(self.clustering_method)].keys() and
                        'depth_1' in grp['clustering/{}'.format(
                            self.clustering_method)].keys()
                        ):
                    data.cluster0 = torch.tensor(
                        grp['clustering/' + self.clustering_method + '/depth_0'][()], dtype=torch.long)
                    data.cluster1 = torch.tensor(
                        grp['clustering/' + self.clustering_method + '/depth_1'][()], dtype=torch.long)
                else:
                    print('WARNING: no cluster detected')
            else:
                print('WARNING: no cluster detected')
        else:
            print('WARNING: no cluster detected')

        f5.close()
        return data

    def create_index_molecules(self):
        '''Creates the indexing of each molecule in the dataset.

        Creates the indexing: [ ('1ak4.hdf5,1AK4_100w),...,('1fqj.hdf5,1FGJ_400w)]
        This allows to refer to one complex with its index in the list
        '''
        print("   Processing data set")

        self.index_complexes = []

        desc = '{:25s}'.format('   Train dataset')
        if self.tqdm:
            data_tqdm = tqdm(
                self.database, desc=desc, file=sys.stdout)
        else:
            print('   Train dataset')
            data_tqdm = self.database
        sys.stdout.flush()

        for fdata in data_tqdm:
            if self.tqdm:
                data_tqdm.set_postfix(mol=os.path.basename(fdata))
            try:
                fh5 = h5py.File(fdata, 'r')
                if self.index is None:
                    mol_names = list(fh5.keys())
                else:
                    mol_names = [list(fh5.keys())[i]
                                 for i in self.index]
                for k in mol_names:
                    if self.filter(fh5[k]):
                        self.index_complexes += [(fdata, k)]
                fh5.close()
            except Exception as inst:
                print('\t\t--> Ignore File : ' + fdata)
                print(inst)

        self.ntrain = len(self.index_complexes)
        self.index_train = list(range(self.ntrain))
        self.ntot = len(self.index_complexes)

    def filter(self, molgrp):
        '''Filters the molecule according to a dictionary.

        The filter is based on the attribute self.dict_filter
        that must be either of the form: { 'name' : cond } or None

        Args:
            molgrp (str): group name of the molecule in the hdf5 file
        Returns:
            bool: True if we keep the complex False otherwise
        Raises:
            ValueError: If an unsuported condition is provided
        '''
        if self.dict_filter is None:
            return True

        for cond_name, cond_vals in self.dict_filter.items():

            try:
                val = molgrp[cond_name][()]
            except KeyError:
                print('   :Filter %s not found for mol %s' %
                      (cond_name, mol))

            # if we have a string it's more complicated
            if isinstance(cond_vals, str):

                ops = ['>', '<', '==']
                new_cond_vals = cond_vals
                for o in ops:
                    new_cond_vals = new_cond_vals.replace(o, 'val'+o)
                if not eval(new_cond_vals):
                    return False
            else:
                raise ValueError(
                    'Conditions not supported', cond_vals)

        return True



import sys
import h5py
import pandas as pd
import numpy as np
 
def hdf5_to_csv(hdf5_path):
        
        hdf5 = h5py.File(hdf5_path,'r+')
        name = hdf5_path.split('.')[0]
        
        first = True
        for epoch in hdf5.keys():
                for dataset in hdf5['{}'.format(epoch)].keys():
                        mol = hdf5['{}/{}/mol'.format(epoch, dataset)]
                        epoch_lst = [epoch] * len(mol)
                        dataset_lst = [dataset] * len(mol)

                        outputs = hdf5['{}/{}/outputs'.format(epoch, dataset)]
                        targets = hdf5['{}/{}/targets'.format(epoch, dataset)]
                        if len(targets) == 0:
                                targets = 'n'*len(mol)

                        # This section is specific to the binary class
                        # it adds the raw output, i.e. probabilities to belong to the class 0 and the class 1, to the prediction hdf5
                        # This way, binary information can be transformed back to continuous data and used for ranking
                        if 'raw_output' in hdf5['{}/{}'.format(epoch, dataset)].keys():
                                if first :
                                        header = ['epoch', 'set', 'model', 'targets', 'prediction', 'raw_prediction_0', 'raw_prediction_1']
                                        output_file = open('{}.csv'.format(name), 'w')
                                        output_file.write(','+','.join(header)+'\n')
                                        output_file.close()
                                        first = False
                                # probability of getting 0         
                                outputs_0 = hdf5['{}/{}/raw_output'.format(epoch, dataset)][()][:,0]
                                # probability of getting 1
                                outputs_1 = hdf5['{}/{}/raw_output'.format(epoch, dataset)][()][:,1]
                                dataset_df = pd.DataFrame(list(zip(epoch_lst, dataset_lst, mol, targets, outputs, outputs_0, outputs_1)), columns = header)
                        
                        else:
                                if first :
                                        header = ['epoch', 'set', 'model', 'targets', 'prediction']
                                        output_file = open('{}.csv'.format(name), 'w')
                                        output_file.write(','+','.join(header)+'\n')
                                        output_file.close()
                                        first = False
                                dataset_df = pd.DataFrame(list(zip(epoch_lst, dataset_lst, mol, targets, outputs)), columns = header)
                        
                        dataset_df.to_csv('{}.csv'.format(name), mode='a', header=False)
        

if __name__ == "__main__":	
        
	if len(sys.argv) != 2 :
        	print ("""\n
This scripts converts the hdf5 output files of GraphProt into csv files
                
Usage: 
python hdf5_to_csv.py file.hdf5
""")
                
	else: 
                try: 
                        hdf5_path = sys.argv[1]
                        hdf5_to_csv(hdf5_path)
                        
                except:
                        print('Please make sure that your input file if a HDF5 file')


import torch
from torch.nn import Parameter
import torch.nn.functional as F
import torch.nn as nn

from torch_scatter import scatter_add
from torch_scatter import scatter_mean

from torch_geometric.utils import remove_self_loops, add_self_loops, softmax

# torch_geometric import
from torch_geometric.nn.inits import uniform
from torch_geometric.nn import max_pool_x

# deeprank_gnn import
from .community_pooling import get_preloaded_cluster, community_pooling


def add_self_loops_wattr(edge_index, edge_attr, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    dtype, device = edge_index.dtype, edge_index.device
    loop = torch.arange(0, num_nodes, dtype=dtype, device=device)
    loop = loop.unsqueeze(0).repeat(2, 1)
    edge_index = torch.cat([edge_index, loop], dim=1)

    dtype, device = edge_attr.dtype, edge_attr.device
    loop = torch.ones(num_nodes, dtype=dtype, device=device)
    edge_attr = torch.cat([edge_attr, loop])

    return edge_index, edge_attr


class GraphAttention(torch.nn.Module):

    """
    This is a new layer that is similar to the graph attention network but simpler

    z_i =  1 / Ni \Sum_j a_ij * [x_i || x_j] * W + b_i

    || is the concatenation operator: [1,2,3] || [4,5,6] = [1,2,3,4,5,6]
    Ni is the number of neighbor of node i
    \Sum_j runs over the neighbors of node i
    a_ij is the edge attribute between node i and j

    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 bias=True):

        super(GraphAttention, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.weight = Parameter(
            torch.Tensor(2 * in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        size = 2 * self.in_channels
        uniform(size, self.weight)
        uniform(size, self.bias)

    def forward(self, x, edge_index, edge_attr):

        #print('weight : ', torch.sum(self.weight))

        row, col = edge_index
        num_node = len(x)
        edge_attr = edge_attr.unsqueeze(
            -1) if edge_attr.dim() == 1 else edge_attr

        # create edge feature by concatenating node feature
        alpha = torch.cat([x[row], x[col]], dim=-1)

        # multiply the edge features with the fliter
        alpha = torch.mm(alpha, self.weight)

        # multiply each edge features with the corresponding dist
        alpha = edge_attr*alpha

        # scatter the resulting edge feature to get node features
        out = torch.zeros(
            num_node, self.out_channels).to(alpha.device)
        out = scatter_mean(alpha, row, dim=0, out=out)

        # if the graph is undirected and (i,j) and (j,i) are both in
        # the edge_index then we do not need to have that second line
        # or we count everythong twice
        #out = scatter_mean(alpha,col,dim=0,out=out)

        # add the bias
        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.in_channels,
                                   self.out_channels)


class GINet(torch.nn.Module):

    def __init__(self, input_shape, output_shape=1):
        super(GINet, self).__init__()

        self.conv1 = GraphAttention(input_shape, 16)
        self.conv2 = GraphAttention(16, 32)

        self.fc1 = torch.nn.Linear(32, 64)
        self.fc2 = torch.nn.Linear(64, output_shape)

        self.clustering = 'mcl'

    def forward(self, data):

        act = nn.Tanhshrink()
        act = F.relu
        #act = nn.LeakyReLU(0.25)

        # first conv block
        data.x = act(self.conv1(
            data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster0, data.batch)
        data = community_pooling(cluster, data)

        # second conv block
        data.x = act(self.conv2(
            data.x, data.edge_index, data.edge_attr))
        cluster = get_preloaded_cluster(data.cluster1, data.batch)
        x, batch = max_pool_x(cluster, data.x, data.batch)

        # FC
        x = scatter_mean(x, batch, dim=0)
        x = act(self.fc1(x))
        x = self.fc2(x)
        #x = F.dropout(x, training=self.training)

        return x
        # return F.relu(x)



import glob
import h5py
import sys
import numpy as np 

def add_target(graph_path, target_name, target_list, sep=' '):

    target_dict = {}

    # Converts the target list into a dictionary
    # The input target list should respect the following format :
    # 1ATN_xxx-1 0
    # 1ATN_xxx-2 1
    # 1ATN_xxx-3 0
    # 1ATN_xxx-4 0

    labels = np.loadtxt(target_list, delimiter=sep, usecols=[0], dtype=np.str)
    values = np.loadtxt(target_list, delimiter=sep, usecols=[1])
    for label, value in zip(labels, values):
        target_dict[label] = value

    first_model = True

    for hdf5 in glob.glob('{}/*.hdf5'.format(graph_path)):
        try:
            f5 = h5py.File(hdf5, 'a')

            for model in f5.keys():
                try :
                    model_gp = f5['{}'.format(model)]

                    if 'score' not in model_gp :
                        model_gp.create_group('score')

                    group=f5['{}/score/'.format(model)]

                    if first_model is True :
                        print('The target {} does not exist. Create it'.format(target_name))
                        first_model = False

                    if target_name in group.keys():
                        # Delete the target if it already existed                                                                                                                    
                        del group[target_name]

                    # Create the target                                                                                                                                              
                    group[target_name] = target_dict[model]

                except:
                    print('no graph for {}'.format(model))
                    
            f5.close()

        except:
            print('no graph for {}'.format(hdf5))   
            



import os
import numpy as np

import pdb2sql
from .FeatureClass import FeatureClass


class AtomicFeature(FeatureClass):

    def __init__(self, pdbfile, param_charge=None, param_vdw=None, patch_file=None,
                 contact_distance=8.5, root_export='./', individual_directory=False, verbose=False):
        '''
        Compute the Coulomb, van der Waals interaction and charges

        Args:

            pdbfile (str): pdb file of the molecule

            param_charge (str): file name of the force field fiel containing the charges e.g. protein-allhdg5.4_new.top. Must be of the format:

                * CYM  atom O   type=O      charge=-0.500 end
                * ALA    atom N   type=NH1     charge=-0.570 end

            param_vdw (str): file name of the force field file containing the vdw parameters e.g  protein-allhdg5.4_new.param. Must be of the format

                * NONBonded  CYAA    0.105   3.750       0.013    3.750
                * NONBonded  CCIS    0.105   3.750       0.013    3.750


            patch_file (str): file name of a valid patch file for the parameters e.g. patch.top
                         The way we handle the patching is very manual and should be
                         made more automatic

            contact_distance (float): the maximum distance between 2 contact atoms

            include_entire_residue (bool): if atom X of resiue M is a contact atom, include all the atoms
                                     of resiude M in the contact_atom list

            root_export (str): root directory where the feature file will be exported (deprecated ?)


        Example :

        >>> pdb = '1AK4_100w.pdb'
        >>>
        >>> # get the force field included in deeprank
        >>> # if another FF has been used to compute the ref
        >>> # change also this path to the correct one
        >>> FF = pkg_resources.resource_filename('deeprank.features','') + '/forcefield/'
        >>>
        >>> # declare the feature calculator instance
        >>> atfeat = AtomicFeature(pdb,
        >>>                        param_charge = FF + 'protein-allhdg5-4_new.top',
        >>>                        param_vdw    = FF + 'protein-allhdg5-4_new.param',
        >>>                        patch_file   = FF + 'patch.top')
        >>> # assign parameters
        >>> atfeat.assign_parameters()
        >>>
        >>> # only compute the pair interactions here
        >>> atfeat.evaluate_pair_interaction(save_interactions=test_name)
        >>>
        >>> # close the db
        >>> atfeat.sqldb.close()
        '''

        super().__init__("Atomic")

        # set a few things
        self.pdbfile = pdbfile
        self.param_charge = param_charge
        self.param_vdw = param_vdw
        self.patch_file = patch_file
        self.contact_distance = contact_distance
        self.individual_directory = individual_directory
        self.verbose = verbose

        # a few constant
        self.eps0 = 1
        self.c = 332.0636

        # dircetory to export
        self.root_export = root_export

        # read the pdb as an sql
        self.sqldb = pdb2sql(self.pdbfile)

        # read the force field
        self.read_charge_file()

        if patch_file != None:
            self.patch = self.read_patch()
        else:
            self.patch = None

        # read the vdw param file
        self.read_vdw_file()

        # get the contact atoms
        self.get_contact_atoms()

    #####################################################################################
    #
    #   READ INPUT FILES
    #
    #####################################################################################

    def read_charge_file(self):
        '''Read the .top file given in entry.

        This function creates :

            - self.charge : dictionary  {(resname,atname):charge}
            - self.valid_resnames : list ['VAL','ALP', .....]
            - self.at_name_type_convertor : dictionary {(resname,atname):attype}

        '''

        f = open(self.param_charge)
        data = f.readlines()
        f.close()

        # loop over all the data
        self.charge = {}
        self.at_name_type_convertor = {}
        resnames = []

        # loop over the file
        for l in data:

            # split the line
            words = l.split()

            # get the resname/atname
            res, atname = words[0], words[2]

            # get the charge
            ind = l.find('charge=')
            q = float(l[ind+7:ind+13])

            # get the type
            attype = words[3].split('=')[-1]

            # store the charge
            self.charge[(res, atname)] = q

            # put the resname in a list so far
            resnames.append(res)

            # dictionary for conversion name/type
            self.at_name_type_convertor[(res, atname)] = attype

        self.valid_resnames = list(set(resnames))

    def read_patch(self):
        '''Read the patchfile.

        This function creates

            - self.patch_charge : Dicitionary   {(resName,atName) : charge}
            - self.patch_type   : Dicitionary   {(resName,atName) : type}

        '''

        f = open(self.patch_file)
        data = f.readlines()
        f.close()

        self.patch_charge, self.patch_type = {}, {}

        for l in data:

            # ignore comments
            if l[0] != '#' and l[0] != '!' and len(l.split()) > 0:

                words = l.split()

                # get the new charge
                ind = l.find('CHARGE=')
                q = float(l[ind+7:ind+13])
                self.patch_charge[(words[0], words[3])] = q

                # get the new type if any
                ind = l.find('TYPE=')
                if ind != -1:
                    type_ = l[ind+5:ind+9]
                    self.patch_type[(words[0], words[3])
                                    ] = type_.strip()

    def read_vdw_file(self):
        ''' Read the .param file

        The patch file must be of the form:

            NONBONDED ATNAME 0.10000 3.298765 0.100000 3.089222

            - First two numbers are for inter-chain interations
            - Last two nmbers are for intra-chain interactions
              (We only compute the interchain here)

        This function creates

            - self.vdw : dictionary {attype:[E1,S1]}
        '''

        f = open(self.param_vdw)
        data = f.readlines()
        f.close()

        self.vdw_param = {}

        for line in data:

            # split the atom
            line = line.split()

            # empty line
            if len(line) == 0:
                continue

            # comment
            if line[0][0] == '#':
                continue

            self.vdw_param[line[1]] = list(map(float, line[2:4]))

    def get_contact_atoms(self):
        """Get the contact atoms only select amino acids.

        The ligands are not considered.
        """

        # position of the chains
        xyz1 = np.array(self.sqldb.get('x,y,z', chainID='A'))
        xyz2 = np.array(self.sqldb.get('x,y,z', chainID='B'))

        # rowID of the second chain
        index_b = self.sqldb.get('rowID', chainID='B')

        # resName of the chains
        resName1 = np.array(self.sqldb.get('resName', chainID='A'))
        resName2 = np.array(self.sqldb.get('resName', chainID='B'))

        # declare the contact atoms
        self.contact_atoms_A = []
        self.contact_atoms_B = []

        # The contact atom pairs only co ntains pairs of atoms that are
        # in contact
        self.contact_pairs = {}

        for i, x0 in enumerate(xyz1):

            # compute the contact atoms
            contacts = np.where(
                np.sqrt(np.sum((xyz2-x0)**2, 1)) < self.contact_distance)[0]

            # if we have contact atoms and resA is not a ligand
            if (len(contacts) > 0) and (resName1[i] in self.valid_resnames):

                # add i to the list
                # add the index of b if its resname is not a ligand
                self.contact_atoms_A += [i]
                self.contact_atoms_B += [index_b[k]
                                         for k in contacts if resName2[k] in self.valid_resnames]

                # add the contact pairs to the list
                self.contact_pairs[i] = [
                    index_b[k] for k in contacts if resName2[k] in self.valid_resnames]

        # create a set of unique indexes
        self.contact_atoms_A = sorted(set(self.contact_atoms_A))
        self.contact_atoms_B = sorted(set(self.contact_atoms_B))

        # if no atoms were found
        if len(self.contact_atoms_A) == 0:
            print('Warning : No contact atoms detected in atomicFeature')

    def _extend_contact_to_residue(self):
        """Extend the contact atoms to entire residue where one atom is contacting."""

        # extract the data
        dataA = self.sqldb.get(
            'chainId,resName,resSeq', rowID=self.contact_atoms_A)
        dataB = self.sqldb.get(
            'chainId,resName,resSeq', rowID=self.contact_atoms_B)

        # create tuple cause we want to hash through it
        dataA = [tuple(x) for x in dataA]
        dataB = [tuple(x) for x in dataB]

        # extract uniques
        resA = list(set(dataA))
        resB = list(set(dataB))

        # init the list
        index_contact_A, index_contact_B = [], []

        # contact of chain A
        for resdata in resA:
            chainID, resName, resSeq = resdata
            index_contact_A += self.sqldb.get(
                'rowID', chainID=chainID, resName=resName, resSeq=resSeq)

        # contact of chain B
        for resdata in resB:
            chainID, resName, resSeq = resdata
            index_contact_B += self.sqldb.get(
                'rowID', chainID=chainID, resName=resName, resSeq=resSeq)

        # make sure that we don't have double (maybe optional)
        index_contact_A = sorted(set(index_contact_A))
        index_contact_B = sorted(set(index_contact_B))

        return index_contact_A, index_contact_B

    #####################################################################################
    #
    #   Assign parameters
    #
    #####################################################################################

    def assign_parameters(self):
        '''Assign to each atom in the pdb its charge and vdw interchain parameters.

        Directly deals with the patch so that we don't loop over the residues
        multiple times
        '''

        # get all the resnumbers
        if self.verbose:
            print('-- Assign force field parameters')

        data = self.sqldb.get('chainID,resSeq,resName')
        natom = len(data)
        data = np.unique(np.array(data), axis=0)

        # declare the parameters for future insertion in SQL
        atcharge = np.zeros(natom)
        ateps = np.zeros(natom)
        atsig = np.zeros(natom)

        # check
        attype = np.zeros(natom, dtype='<U5')
        ataltResName = np.zeros(natom, dtype='<U5')

        # add attribute to the db

        # loop over all the residues
        for chain, resNum, resName in data:

            # atom types of the residue
            #query = "WHERE chainID='%s' AND resSeq=%s" %(chain,resNum)
            atNames = np.array(self.sqldb.get(
                'name', chainID=chain, resSeq=resNum))
            rowID = np.array(self.sqldb.get(
                'rowID', chainID=chain, resSeq=resNum))

            # get the alternative resname
            altResName = self._get_altResName(resName, atNames)

            # get the charge of this residue
            atcharge[rowID] = self._get_charge(
                resName, altResName, atNames)

            # get the vdw parameters
            eps, sigma, type_ = self._get_vdw(
                resName, altResName, atNames)
            ateps[rowID] += eps
            atsig[rowID] += sigma

            ataltResName[rowID] = altResName
            attype[rowID] = type_

        # put the charge in SQL
        self.sqldb.add_column('CHARGE')
        self.sqldb.update_column('CHARGE', atcharge)

        # put the VDW in SQL
        self.sqldb.add_column('eps')
        self.sqldb.update_column('eps', ateps)

        self.sqldb.add_column('sig')
        self.sqldb.update_column('sig', atsig)

        self.sqldb.add_column('type', 'TEXT')
        self.sqldb.update_column('type', attype)

        self.sqldb.add_column('altRes', 'TEXT')
        self.sqldb.update_column('altRes', ataltResName)

    @staticmethod
    def _get_altResName(resName, atNames):
        ''' Apply the patch data.

        This is adopted from preScan.pl
        This is very static and I don't quite like it
        The structure of the dictionary is as following

        { NEWRESTYPE : [ 'OLDRESTYPE' , [atom types that must be present], [atom types that must NOT be present] }   ]  }

        Args:
            resName (str): name of the residue
            atNames (list(str)): names of the atoms
        '''

        new_type = {
            'PROP': ['all',    ['HT1', 'HT2'],                    []],
            'NTER': ['all',    ['HT1', 'HT2', 'HT3'],              []],
            'CTER': ['all',    ['OXT'],                          []],
            'CTN': ['all',    ['NT', 'HT1', 'HT2'],               []],
            'CYNH': ['CYS',    ['1SG'],                          ['2SG']],
            'DISU': ['CYS',    ['1SG', '2SG'],                    []],
            'HISE': ['HIS',    ['ND1', 'CE1', 'CD2', 'NE2', 'HE2'],  ['HD1']],
            'HISD': ['HIS',    ['ND1', 'CE1', 'CD2', 'NE2', 'HD1'],  ['HE2']]
        }

        # this works fine now

        altResName = resName
        for key, values in new_type.items():

            res, atpres, atabs = values

            if res == resName or res == 'all':
                if all(x in atNames for x in atpres) and all(x not in atNames for x in atabs):
                    altResName = key

        return altResName

    def _get_vdw(self, resName, altResName, atNames):
        """Get vdw itneraction terms.

        Args:
            resName (str): name of the residue
            altResName (str): alternative name of the residue
            atNames (list(str)): names of the atoms
        """

        # in case the resname is not valid
        if resName not in self.valid_resnames:
            vdw_eps = [0.00]*len(atNames)
            vdw_sigma = [0.00]*len(atNames)
            type_ = ['None']*len(atNames)

            return vdw_eps, vdw_sigma, type_

        vdw_eps, vdw_sigma, type_ = [], [], []

        for at in atNames:

            if (altResName, at) in self.patch_type:
                type_.append(self.patch_type[(altResName, at)])
                vdw_data = self.vdw_param[self.patch_type[(
                    altResName, at)]]
                vdw_eps.append(vdw_data[0])
                vdw_sigma.append(vdw_data[1])

            elif (resName, at) in self.at_name_type_convertor:
                type_.append(
                    self.at_name_type_convertor[(resName, at)])
                vdw_data = self.vdw_param[self.at_name_type_convertor[(
                    resName, at)]]
                vdw_eps.append(vdw_data[0])
                vdw_sigma.append(vdw_data[1])

            else:
                type_.append('None')
                #print('Warning : atom type %s not found for resType %s or patch type %s' %(at,resName,altResName))
                vdw_eps.append(0.0)
                vdw_sigma.append(0.0)

        return vdw_eps, vdw_sigma, type_

    def _get_charge(self, resName, altResName, atNames):
        """Get the charge information.

        Args:
            resName (str): name of the residue
            altResName (str): alternative name of the residue
            atNames (list(str)): names of the atoms
        """

        # in case the resname is not valid
        if resName not in self.valid_resnames:
            q = [0.0]*len(atNames)
            return q

        # assign the charges
        q = []
        for at in atNames:
            if (altResName, at) in self.patch_charge:
                q.append(self.patch_charge[(altResName, at)])

            elif (resName, at) in self.charge:
                q.append(self.charge[(resName, at)])

            else:
                q.append(0.0)

        return q

    #####################################################################################
    #
    #   Simple charges
    #
    #####################################################################################

    def evaluate_charges(self, extend_contact_to_residue=False):
        """ Evaluate the charges.

        Args:
            extend_contact_to_residue (bool, optional): extend to res

        """
        if self.verbose:
            print('-- Compute list charge for contact atoms only')

        if len(self.contact_atoms_A) == 0:
            self.feature_data['charge'] = {}
            self.feature_data_xyz['charge'] = {}
            self.export_directories['charge'] = self.root_export+'/CHARGE/'
            return

        # extract information from the pdb2sq
        xyz = np.array(self.sqldb.get('x,y,z'))
        atinfo = self.sqldb.get('chainID,resName,resSeq,name')

        charge = np.array(self.sqldb.get('CHARGE'))

        # define the dictionaries
        charge_data = {}
        charge_data_xyz = {}

        # entire residue or not
        if extend_contact_to_residue:
            indA, indB = self._extend_contact_to_residue()
            index_contact_atoms = indA+indB
        else:
            index_contact_atoms = self.contact_atoms_A + self.contact_atoms_B

        # loop over the chain A
        for i in index_contact_atoms:

            # atinfo
            key = tuple(atinfo[i])

            # store in the dicts
            charge_data[key] = [charge[i]]

            # xyz format
            chain_dict = [{'A': 0, 'B': 1}[key[0]]]
            key = tuple(chain_dict + xyz[i, :].tolist())
            charge_data_xyz[key] = [charge[i]]

        # if we have no contact atoms
        if len(charge_data_xyz) == 0:
            charge_data_xyz[tuple([0, 0., 0., 0.])] = [0.0]
            charge_data_xyz[tuple([1, 0., 0., 0.])] = [0.0]

        # add the electrosatic feature
        self.feature_data['charge'] = charge_data
        self.feature_data_xyz['charge'] = charge_data_xyz

        # is that obsolte ?
        if self.individual_directory:
            self.export_directories['charge'] = self.root_export+'/CHARGE/'
        else:
            self.export_directories['charge'] = self.root_export

    #####################################################################################
    #
    #   PAIR INTERACTIONS
    #
    #####################################################################################

    def evaluate_pair_interaction(self, print_interactions=False, save_interactions=False):
        """Evalaute the pair interactions (coulomb and vdw).

        Args:
            print_itneractions (bool, optional): print data to screen
            save_interactions (bool, optional): save the itneractions to file.
        """

        if self.verbose:
            print('-- Compute interaction energy for contact pairs only')

        if len(self.contact_atoms_A) == 0:
            self.feature_data['coulomb'] = {}
            self.feature_data_xyz['culomb'] = {}
            self.export_directories['coulomb'] = self.root_export+'/ELEC/'
            return

        # extract information from the pdb2sq
        xyz = np.array(self.sqldb.get('x,y,z'))
        atinfo = self.sqldb.get('chainID,resName,resSeq,name')

        charge = np.array(self.sqldb.get('CHARGE'))
        vdw = np.array(self.sqldb.get('eps,sig'))
        eps, sig = vdw[:, 0], vdw[:, 1]

        # define the dictionaries
        # these holds data like
        # chainID resname resSeq,name values
        electro_data = {}
        vdw_data = {}

        # define the dict that hold
        #  x y z values
        electro_data_xyz = {}
        vdw_data_xyz = {}

        # define the matrices
        natA, natB = len(self.sqldb.get('x', chainID='A')), len(
            self.sqldb.get('x', chainID='B'))
        matrix_elec = np.zeros((natA, natB))
        matrix_vdw = np.zeros((natA, natB))

        # handle the export of the interaction breakdown
        _save_ = False
        if save_interactions != False:
            if save_interactions == True:
                save_interactions = './'
            if os.path.isdir(save_interactions):
                fname = save_interactions + '/atomic_pair_interaction.dat'
            else:
                fname = save_interactions
            f = open(fname, 'w')
            _save_ = True

        # total energy terms
        ec_tot, evdw_tot = 0, 0

        # loop over the chain A
        for iA, indsB in self.contact_pairs.items():

            # coulomb terms
            r = np.sqrt(np.sum((xyz[indsB, :]-xyz[iA, :])**2, 1))
            q1q2 = charge[iA]*charge[indsB]
            ec = q1q2 * self.c / (self.eps0*r) * \
                (1 - (r/self.contact_distance)**2) ** 2

            # coulomb terms
            sigma_avg = 0.5*(sig[iA] + sig[indsB])
            eps_avg = np.sqrt(eps[iA]*eps[indsB])

            # normal LJ potential
            evdw = 4.0 * eps_avg * \
                ((sigma_avg/r)**12 - (sigma_avg/r)**6) * \
                self._prefactor_vdw(r)

            # total energy terms
            ec_tot += np.sum(ec)
            evdw_tot += np.sum(evdw)

            # atinfo
            keyA = tuple(atinfo[iA])

            # store in matrix form so that
            # we don't have to recalculate for B
            indb_matrix = [i - natA for i in indsB]
            matrix_elec[iA, indb_matrix] = ec
            matrix_vdw[iA, indb_matrix] = evdw

            # store in the dicts
            electro_data[keyA] = [np.sum(ec)]
            vdw_data[keyA] = [np.sum(evdw)]

            # store in the xyz dict
            key = tuple([0]+xyz[iA, :].tolist())
            electro_data_xyz[key] = [np.sum(ec)]
            vdw_data_xyz[key] = [np.sum(evdw)]

            # print the result
            if _save_ or print_interactions:

                for iB, indexB in enumerate(indsB):

                    line = ''
                    keyB = tuple(atinfo[indexB])

                    line += '{:<3s}'.format(keyA[0])
                    line += '\t{:>1s}'.format(keyA[1])
                    line += '\t{:>4d}'.format(keyA[2])
                    line += '\t{:^4s}'.format(keyA[3])

                    line += '\t{:<3s}'.format(keyB[0])
                    line += '\t{:>1s}'.format(keyB[1])
                    line += '\t{:>4d}'.format(keyB[2])
                    line += '\t{:^4s}'.format(keyB[3])

                    line += '\t{: 6.3f}'.format(r[iB])
                    line += '\t{: f}'.format(ec[iB])
                    line += '\t{: e}'.format(evdw[iB])

                    # print and/or save the interactions
                    if print_interactions:
                        print(line)

                    if _save_:
                        line += '\n'
                        f.write(line)

        # print the total interactions
        if print_interactions or _save_:
            line = '\n\n'
            line += 'Total Evdw  = {:> 12.8f}\n'.format(evdw_tot)
            line += 'Total Eelec = {:> 12.8f}\n'.format(ec_tot)
            if print_interactions:
                print(line)
            if _save_:
                f.write(line)

        # close export file
        if _save_:
            f.close()

        # loop over the B atoms
        for indexB in self.contact_atoms_B:

            # atinfo
            keyB = tuple(atinfo[indexB])

            # extract the values from the matrix
            ec = matrix_elec[:, indexB-natA]
            evdw = matrix_vdw[:, indexB-natA]

            # store in the dict
            electro_data[keyB] = [np.sum(ec)]
            vdw_data[keyB] = [np.sum(evdw)]

            # store the xyz dict
            key = tuple([1]+xyz[indexB, :].tolist())
            electro_data_xyz[key] = [np.sum(ec)]
            vdw_data_xyz[key] = [np.sum(evdw)]

        # if we have no contact atoms
        if len(electro_data_xyz) == 0:
            electro_data_xyz0[tuple([0, 0., 0., 0.])] = [0.0]
            electro_data_xyz0[tuple([1, 0., 0., 0.])] = [0.0]

        if len(vdw_data_xyz) == 0:
            vdw_data_xyz0[tuple([0, 0., 0., 0.])] = [0.0]
            vdw_data_xyz0[tuple([1, 0., 0., 0.])] = [0.0]

        # add the electrosatic feature
        self.feature_data['coulomb'] = electro_data
        self.feature_data_xyz['coulomb'] = electro_data_xyz

        # add the vdw feature
        self.feature_data['vdwaals'] = vdw_data
        self.feature_data_xyz['vdwaals'] = vdw_data_xyz

        # dict for export
        # I think that's obsolete
        if self.individual_directory:
            self.export_directories['coulomb'] = self.root_export+'/ELEC/'
        else:
            self.export_directories['coulomb'] = self.root_export

        if self.individual_directory:
            self.export_directories['vdwaals'] = self.root_export+'/VDW/'
        else:
            self.export_directories['vdwaals'] = self.root_export

    #####################################################################################
    #
    #   ELECTROSTATIC
    #
    #####################################################################################

    def compute_coulomb_interchain_only(self, dosum=True, contact_only=False):
        """Compute the coulomb interactions between the chains only.

        Args:
            dosum (bool, optional): sum the interaction terms for each atoms
            contact_only (bool, optional): consider only contact atoms
        """

        if self.verbose:
            print('-- Compute coulomb energy interchain only')

        if contact_only:

            if len(self.contact_atoms_A) == 0:
                self.feature_data['coulomb'] = {}
                self.export_directories['coulomb'] = self.root_export+'/ELEC/'
                return

            xyzA = np.array(self.sqldb.get(
                'x,y,z', rowID=self.contact_atoms_A))
            xyzB = np.array(self.sqldb.get(
                'x,y,z', rowID=self.contact_atoms_B))

            chargeA = np.array(self.sqldb.get(
                'CHARGE', rowID=self.contact_atoms_A))
            chargeB = np.array(self.sqldb.get(
                'CHARGE', rowID=self.contact_atoms_B))

            atinfoA = self.sqldb.get(
                'chainID,resName,resSeq,name', rowID=self.contact_atoms_A)
            atinfoB = self.sqldb.get(
                'chainID,resName,resSeq,name', rowID=self.contact_atoms_B)

        else:

            xyzA = np.array(self.sqldb.get('x,y,z', chainID='A'))
            xyzB = np.array(self.sqldb.get('x,y,z', chainID='B'))

            chargeA = np.array(self.sqldb.get('CHARGE', chainID='A'))
            chargeB = np.array(self.sqldb.get('CHARGE', chainID='B'))

            atinfoA = self.sqldb.get(
                'chainID,resName,resSeq,name', chainID='A')
            atinfoB = self.sqldb.get(
                'chainID,resName,resSeq,name', chainID='B')

        natA, natB = len(xyzA), len(xyzB)
        matrix = np.zeros((natA, natB))

        electro_data = {}

        for iat in range(natA):

            # coulomb terms
            r = np.sqrt(np.sum((xyzB-xyzA[iat, :])**2, 1))
            q1q2 = chargeA[iat]*chargeB
            value = q1q2/r

            # store amd symmtrized these values
            matrix[iat, :] = value

            # atinfo
            key = tuple(atinfoA[iat])

            # store
            if dosum:
                value = [np.sum(value)]
            electro_data[key] = value

        for iat in range(natB):

            # atinfo
            key = tuple(atinfoB[iat])

            # store
            value = matrix[:, iat]
            if dosum:
                value = [np.sum(value)]
            electro_data[key] = value

        # add the feature to the dictionary of features
        self.feature_data['coulomb'] = electro_data
        if self.individual_directory:
            self.export_directories['coulomb'] = self.root_export+'/ELEC/'
        else:
            self.export_directories['coulomb'] = self.root_export

    #####################################################################################
    #
    #   VAN DER WAALS
    #
    #####################################################################################
    def compute_vdw_interchain_only(self, dosum=True, contact_only=False):
        """Compute the vdw interactions between the chains only.

        Args:
            dosum (bool, optional): sum the interaction terms for each atoms
            contact_only (bool, optional): consider only contact atoms
        """
        if self.verbose:
            print('-- Compute vdw energy interchain only')

        if contact_only:

            if len(self.contact_atoms_A) == 0:
                self.feature_data['coulomb'] = {}
                self.export_directories['coulomb'] = self.root_export+'/ELEC/'
                return
            xyzA = np.array(self.sqldb.get(
                'x,y,z', rowID=self.contact_atoms_A))
            xyzB = np.array(self.sqldb.get(
                'x,y,z', rowID=self.contact_atoms_B))

            vdwA = np.array(self.sqldb.get(
                'eps,sig', rowID=self.contact_atoms_A))
            vdwB = np.array(self.sqldb.get(
                'eps,sig', rowID=self.contact_atoms_B))

            epsA, sigA = vdwA[:, 0], vdwA[:, 1]
            epsB, sigB = vdwB[:, 0], vdwB[:, 1]

            atinfoA = self.sqldb.get(
                'chainID,resName,resSeq,name', rowID=self.contact_atoms_A)
            atinfoB = self.sqldb.get(
                'chainID,resName,resSeq,name', rowID=self.contact_atoms_B)

        else:

            xyzA = np.array(self.sqldb.get('x,y,z', chainID='A'))
            xyzB = np.array(self.sqldb.get('x,y,z', chainID='B'))

            vdwA = np.array(self.sqldb.get('eps,sig', chainID='A'))
            vdwB = np.array(self.sqldb.get('eps,sig', chainID='B'))

            epsA, sigA = vdwA[:, 0], vdwA[:, 1]
            epsB, sigB = vdwB[:, 0], vdwB[:, 1]

            atinfoA = self.sqldb.get(
                'chainID,resName,resSeq,name', chainID='A')
            atinfoB = self.sqldb.get(
                'chainID,resName,resSeq,name', chainID='B')

        natA, natB = len(xyzA), len(xyzB)
        matrix = np.zeros((natA, natB))

        vdw_data = {}

        for iat in range(natA):

            # coulomb terms
            r = np.sqrt(np.sum((xyzB-xyzA[iat, :])**2, 1))
            sigma = 0.5*(sigA[iat] + sigB)
            eps = np.sqrt(epsA[iat]*epsB)

            # normal LJ potential
            value = 4*eps * ((sigma/r)**12 - (sigma/r)**6)

            # store these values
            matrix[iat, :] = value

            # atinfo
            key = tuple(atinfoA[iat])

            # store
            if dosum:
                value = [np.sum(value)]
            vdw_data[key] = value

        for iat in range(natB):

            # atinfo
            key = tuple(atinfoB[iat])

            # store
            value = matrix[:, iat]
            if dosum:
                value = [np.sum(value)]
            vdw_data[key] = value

        # add the feature to the dictionary of features
        self.feature_data['vdwaals'] = vdw_data
        if self.individual_directory:
            self.export_directories['vdwaals'] = self.root_export+'/VDW/'
        else:
            self.export_directories['vdwaals'] = self.root_export

    @staticmethod
    def _prefactor_vdw(r):
        """prefactor for vdw interactions."""

        r_off, r_on = 8.5, 6.5
        r2 = r**2
        pref = (r_off**2-r2)**2 * (r_off**2 - r2 - 3 *
                                   (r_on**2 - r2)) / (r_off**2-r_on**2)**3
        pref[r > r_off] = 0.
        pref[r < r_on] = 1.0
        return pref


#####################################################################################
#
#   THE MAIN FUNCTION CALLED IN THE INTERNAL FEATURE CALCULATOR
#
#####################################################################################

def __compute_feature__(pdb_data, featgrp, featgrp_raw):
    """Main function called in deeprank for the feature calculations

    Args:
        pdb_data (list(bytes)): pdb information
        featgrp (str): name of the group where to save xyz-val data
        featgrp_raw (str): name of the group where to save human readable data
    """
    path = os.path.dirname(os.path.realpath(__file__))
    FF = path + '/forcefield/'

    atfeat = AtomicFeature(pdb_data,
                           param_charge=FF + 'protein-allhdg5-4_new.top',
                           param_vdw=FF + 'protein-allhdg5-4_new.param',
                           patch_file=FF + 'patch.top')

    atfeat.assign_parameters()

    # only compute the pair interactions here
    atfeat.evaluate_pair_interaction(print_interactions=False)

    # compute the charges
    # here we extand the contact atoms to
    # entire residue containing at least 1 contact atom
    atfeat.evaluate_charges(extend_contact_to_residue=True)

    # export in the hdf5 file
    atfeat.export_dataxyz_hdf5(featgrp)
    atfeat.export_data_hdf5(featgrp_raw)

    # close
    atfeat.sqldb.close()



import sys
import os
import torch
import numpy as np

from sklearn import metrics
from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_auc_score

def get_binary(values, threshold, target):
    """
    Transforms continuous or multiclass values into binary values (0/1)

    Args:
        values (list): vector of the target values
        threshold (int or float): threshold used to assign a binary value
                                0 is assigned to 'bad' values;
                                1 is assigned to 'good' values
        target (string): target (y)
                        if target is 'fnat' or 'bin_class': target value > threshold = 1
                        esle: target value > threshold = 0

    Returns:
        list: list of binary values
    """
    inverse = ['fnat', 'bin_class']
    if target in inverse:
        values_binary = [1 if x > threshold else 0 for x in values]
    else:
        values_binary = [1 if x < threshold else 0 for x in values]

    return values_binary


def get_comparison(prediction, ground_truth, binary=True, classes=[0, 1]):
    """
    Computes the confusion matrix to get the number of:

    - false positive (FP)
    - false negative (FN)
    - true positive (TP)
    - true negative (TN)

    Args:
        prediction (list): list of predicted values
        ground_truth (list): list of target values (y)
        binary (bool, optional): If binary is True, the function will return a single value for each FP/FN/TP/TN variable.
                                 If binary is False, the function will return a vector of n values for each FP/FN/TP/TN
                                 Defaults to True.
        classes (list, optional): Array-like of shape (n_classes). Defaults to [0, 1].

    Returns:
        int: false_positive, false_negative, true_positive, true_negative
    """
    CM = confusion_matrix(ground_truth, prediction, labels=classes)

    false_positive = CM.sum(axis=0) - np.diag(CM)
    false_negative = CM.sum(axis=1) - np.diag(CM)
    true_positive = np.diag(CM)
    true_negative = CM.sum() - (false_positive + false_negative + true_positive)

    if binary == True:
        return false_positive[1], false_negative[1], true_positive[1], true_negative[1]

    else:
        return false_positive, false_negative, true_positive, true_negative


class Metrics(object):

    def __init__(self, prediction, y, target, threshold=4, binary=True):
        """
        Master class from which all metrics are computed

        Computed metrics:

        Classification metrics:
        - self.sensitivity: Sensitivity, hit rate, recall, or true positive rate
        - self.specificity: Specificity or true negative rate
        - self.precision: Precision or positive predictive value
        - self.NPV: Negative predictive value
        - self.FPR: Fall out or false positive rate
        - self.FNR: False negative rate
        - self.FDR: False discovery rate
        - self.accuracy: Accuracy

        - self.auc(): AUC
        - self.hitrate(): Hit rate

        Regression metrics:
        - self.explained_variance: Explained variance regression score function
        - self.max_error: Max_error metric calculates the maximum residual error
        - self.mean_abolute_error: Mean absolute error regression loss
        - self.mean_squared_error: Mean squared error regression loss
        - self.root_mean_squared_error: Root mean squared error regression loss
        - self.mean_squared_log_error: Mean squared logarithmic error regression loss
        - self.median_squared_log_error: Median absolute error regression loss
        - self.r2_score: R^2 (coefficient of determination) regression score function

        Args:
            prediction (list): predicted values
            y (list): list of target values
            target (string): irmsd, fnat, capri_class, bin_class
            binary (bool, optional): transform the data in binary vectors. Defaults to True.
            threshold (int, optional): threshold used to split the data into a binary vector. Defaults to 4.
        """
        self.prediction = prediction
        self.y = y
        self.binary = binary
        self.target = target
        self.threshold = threshold

        print('Threshold set to {}'.format(self.threshold))

        if self.binary == True:
            prediction_binary = get_binary(
                self.prediction, self.threshold, self.target)
            y_binary = get_binary(
                self.y, self.threshold, self.target)
            classes = [0, 1]
            false_positive, false_negative, true_positive, true_negative = get_comparison(
                prediction_binary, y_binary, self.binary, classes=classes)

        else:
            if target == 'capri_class':
                classes = [1, 2, 3, 4, 5]
            elif target == 'bin_class':
                classes = [0, 1]
            else:
                raise ValueError('target must be capri_class on bin_class')
            false_positive, false_negative, true_positive, true_negative = get_comparison(
                self.prediction, self.y, self.binary, classes=classes)

        try:
            # Sensitivity, hit rate, recall, or true positive rate
            self.sensitivity = true_positive/(true_positive+false_negative)
        except:
            self.sensitivity = None

        try:
            # Specificity or true negative rate
            self.specificity = true_negative/(true_negative+false_positive)
        except:
            self.specificity = None

        try:
            # Precision or positive predictive value
            self.precision = true_positive/(true_positive+false_positive)
        except:
            self.precision = None

        try:
            # Negative predictive value
            self.NPV = true_negative/(true_negative+false_negative)
        except:
            self.NPV = None

        try:
            # Fall out or false positive rate
            self.FPR = false_positive/(false_positive+true_negative)
        except:
            self.FPR = None

        try:
            # False negative rate
            self.FNR = false_negative/(true_positive+false_negative)
        except:
            self.FNR = None

        try:
            # False discovery rate
            self.FDR = false_positive/(true_positive+false_positive)
        except:
            self.FDR = None

        self.accuracy = (true_positive+true_negative)/(true_positive+false_positive+false_negative+true_negative)

        # regression metrics
        self.explained_variance = None
        self.max_error =  None
        self.mean_abolute_error = None
        self.mean_squared_error = None
        self.root_mean_squared_error = None
        self.mean_squared_log_error = None
        self.median_squared_log_error = None
        self.r2_score = None

        if target in ['fnat', 'irmsd', 'lrmsd']:

            # Explained variance regression score function
            self.explained_variance = metrics.explained_variance_score(self.y, self.prediction)

            # Max_error metric calculates the maximum residual error
            self.max_error = metrics.max_error(self.y, self.prediction)

            # Mean absolute error regression loss
            self.mean_absolute_error = metrics.mean_absolute_error(self.y, self.prediction)

            # Mean squared error regression loss
            self.mean_squared_error = metrics.mean_squared_error(self.y, self.prediction, squared = True)

            # Root mean squared error regression loss
            self.root_mean_squared_error = metrics.mean_squared_error(self.y, self.prediction, squared = False)

            try:
                # Mean squared logarithmic error regression loss
                self.mean_squared_log_error = metrics.mean_squared_log_error(self.y, self.prediction)
            except ValueError:
                print ("WARNING: Mean Squared Logarithmic Error cannot be used when "
                            "targets contain negative values.")

            # Median absolute error regression loss
            self.median_squared_log_error = metrics.median_absolute_error(self.y, self.prediction)

            # R^2 (coefficient of determination) regression score function
            self.r2_score = metrics.r2_score(self.y, self.prediction)


    def format_score(self):
        """
        Sorts the predicted values depending on the target:

        - if target is fnat or bin_class: the highest value the better ranked
        - else: the lowest value the better ranked

        Returns:
            lists: ranks of the predicted values and
                    the corresponding binary (0/1) target values
        """
        idx = np.argsort(self.prediction)

        inverse = ['fnat', 'bin_class']
        if self.target in inverse:
            idx = idx[::-1]

        ground_truth_bool = get_binary(
            self.y, self.threshold, self.target)
        ground_truth_bool = np.array(ground_truth_bool)
        return idx, ground_truth_bool


    def hitrate(self):
        """
        Sorts the target boolean values (0/1) according to the ranks of predicted values

        Returns:
            list: the cumulative sum of hits (1)
        """
        idx, ground_truth_bool = self.format_score()
        return np.cumsum(ground_truth_bool[idx])

    def auc(self):
        """
        Computes the Receiver Operating Characteristic (ROC) area under the curve (AUC)

        Returns:
            float: AUC of the ROC curve
        """
        idx, ground_truth_bool = self.format_score()
        return roc_auc_score(ground_truth_bool, idx)



import numpy as np
import torch
import torch.nn as nn
from torch.nn import MSELoss
import torch.nn.functional as F
from torch_scatter import scatter_mean
from torch_geometric.datasets import MNISTSuperpixels
import torch_geometric.transforms as T
from torch_geometric.data import DataLoader
from torch_geometric.utils import normalized_cut
from torch_geometric.nn import GCNConv, ChebConv, SplineConv, NNConv, GATConv
from torch_geometric.nn import graclus, max_pool, max_pool_x
import matplotlib.pyplot as plt
from tqdm import tqdm

from DataSet import HDF5DataSet
from community_pooling import *

from wgat_conv import WGATConv


index = np.arange(400)
#np.random.shuffle(index)

index_train = index[0:50]
index_test = index[350:]
batch_size = 2



h5train = '1AK4_residue.hdf5'
h5test = '1AK4_residue.hdf5'

target = 'irmsd'
node_feature = ['type','polarity','bsa']
edge_attr = ['dist']

train_dataset = HDF5DataSet(root='./',database=h5train,index=index_train,
                            node_feature=node_feature,edge_feature=edge_attr,
                            target=target)
train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=False)
d = train_dataset.get(0)

test_dataset = HDF5DataSet(root='./',database=h5test,index=index_test,
                           node_feature=node_feature,edge_feature=edge_attr,
                           target=target)
test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)



class Net(torch.nn.Module):
    def __init__(self):
        super(Net, self).__init__()

        self.conv1 = WGATConv(d.num_features, 16)
        self.conv2 = WGATConv(16 , 32)
        #self.conv3 = WGATConv(32 , 32)

        self.fc1 = torch.nn.Linear(32, 64)
        self.fc2 = torch.nn.Linear(64, 1)

    def forward(self, data):

        act = nn.Tanhshrink()
        act = F.relu
        #act = nn.LeakyReLU(0.25)

        data.x = act(self.conv1(data.x, data.edge_index,data.edge_attr))
        cluster = community_detection(data.internal_edge_index,data.num_nodes,edge_attr=None,batches=data.batches)
        data = community_pooling(cluster, data)

        data.x = act(self.conv2(data.x, data.edge_index,data.edge_attr))
        cluster = community_detection(data.internal_edge_index,data.num_nodes,edge_attr=None)
        x, batch = max_pool_x(cluster, data.x, data.batch)

        x = scatter_mean(x, batch, dim=0)
        x = act(self.fc1(x))
        x = self.fc2(x)
        #x = F.dropout(x, training=self.training)

        return x
        #return F.relu(x)


device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
model = Net().to(device)
optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
loss = MSELoss()

def train(epoch):
    model.train()

    if epoch == 50:
        for param_group in optimizer.param_groups:
            param_group['lr'] = 0.01

    if epoch == 75:
        for param_group in optimizer.param_groups:
            param_group['lr'] = 0.001

    for data in (train_loader):
        data = data.to(device)
        optimizer.zero_grad()
        out = model(data).reshape(-1)
        loss(out, data.y).backward()
        optimizer.step()


def test():
    model.eval()
    loss_func = MSELoss()
    loss_val = 0

    for data in test_loader:
        data = data.to(device)
        pred = model(data).reshape(-1)
        loss_val += loss_func(pred,data.y)
    return loss_val


for epoch in range(1, 50):
    train(epoch)
    test_acc = test()
    print('Epoch: {:02d}, Test: {:.4f}'.format(epoch, test_acc))
    #print('Epoch: {:02d}'.format(epoch))

pred, truth = [], []
for data in train_loader:
    data = data.to(device)
    truth += data.y.tolist()
    pred += model(data).reshape(-1).tolist()

test_pred, test_truth = [], []
for data in test_loader:
    data = data.to(device)
    test_truth += data.y.tolist()
    test_pred += model(data).reshape(-1).tolist()
plt.scatter(truth,pred,c='blue')
plt.scatter(test_truth,test_pred,c='red')
#plt.plot([0,1],[0,1])
plt.show()


from sklearn import manifold, datasets

def manifold_embedding(pos,method='tsne'):
    n_components = 2
    n_neighbors = 100

    if method == 'tsne':
        tsne = manifold.TSNE(n_components=n_components, init='pca', random_state=0)
        Y = tsne.fit_transform(pos)
    elif method == 'spectral':
        se = manifold.SpectralEmbedding(n_components=n_components,n_neighbors=n_neighbors)
        Y = se.fit_transform(pos)
    elif method == 'mds':
        mds = manifold.MDS(n_components, max_iter=100, n_init=1)
        Y = mds.fit_transform(pos)
    return Y


import glob
import sys


def pssm_3dcons_to_deeprank(pssm_file):

    pssm = open(pssm_file, 'r').readlines()
    pssm_name = pssm_file.rsplit('.', 1)[0]
    new_pssm = open('{}.deeprank.pssm'.format(pssm_name), 'w')

    firstline=True

    for line in pssm :
       
        if firstline is True: 
            firstline = False
            new_pssm.write('pdbresi pdbresn seqresi seqresn    A    R    N    D    C    Q    E\
    G    H    I    L    K    M    F    P    S    T    W    Y    V   IC\n')
       
        if len(line.split( )) == 44:
            resid = line[0:6].strip()
            resn = line[6]
            pssm_content = line[11:90] 
            ic = line.split( )[-1] 

            new_pssm.write('{0:>5} {1:1} {0:>5} {1:1}    {2} {3}\n'.format(resid, resn, pssm_content,ic))

    new_pssm.close()

if __name__ == "__main__":	
    if len(sys.argv) != 2 :
        print ("""\n
This scripts converts the 3dcons pssm files into deeprank pssm format
                
Usage: 
python 3dcons_to_deeprank_pssm.py [path_to_pssm]
""")

    else: 
        try: 
            pssm_path = sys.argv[1]
            for pssm_file in glob.glob('{}/*.pssm'.format(pssm_path)):
                pssm_3dcons_to_deeprank(pssm_file)
        except:
            print('You must provide the path to the pssm files')

    



import community
import markov_clustering as mc
import networkx as nx
import torch

from torch_scatter import scatter_max, scatter_mean
from torch_geometric.nn.pool.pool import pool_edge, pool_batch, pool_pos
from torch_geometric.nn.pool.consecutive import consecutive_cluster
from torch_geometric.data import Batch, Data


import numpy as np

from time import time

def plot_graph(graph,cluster):

    import matplotlib.pyplot as plt
    pos = nx.spring_layout(graph,iterations=200)
    nx.draw(graph,pos,node_color=cluster)
    plt.show()


def get_preloaded_cluster(cluster,batch):
    
    nbatch = torch.max(batch)+1
    for ib in range(1,nbatch):
        cluster[batch==ib] += torch.max(cluster[batch==ib-1]) + 1
    return cluster


def community_detection_per_batch(edge_index,batch,num_nodes,edge_attr=None,method='mcl'):

    # make the networkX graph
    g = nx.Graph()
    g.add_nodes_from(range(num_nodes))

    for iedge,(i,j) in enumerate(edge_index.transpose(0,1).tolist()):
        if edge_attr is None:
            g.add_edge(i,j)
        else:
            g.add_edge(i,j,weight=edge_attr[iedge])

    num_batch = max(batch)+1
    all_index = range(num_nodes)
    cluster, ncluster = [], 0

    for iB in range(num_batch):

        index = torch.tensor(all_index)[batch==iB].tolist()
        subg = g.subgraph(index)

        # detect the communities using Louvain method
        if method == 'louvain':
            c = community.best_partition(subg)
            cluster += [v+ncluster for k,v in c.items()]
            ncluster = max(cluster)

        # detect communities using MCL
        elif method == 'mcl':
            matrix = nx.to_scipy_sparse_matrix(subg)
            result = mc.run_mcl(matrix)           # run MCL with default parameters
            mc_clust = mc.get_clusters(result)    # get clusters

            index = np.zeros(subg.number_of_nodes()).astype('int')
            for ic, c in enumerate(mc_clust):
                index[list(c)] = ic+ncluster
            cluster += index.tolist()
            ncluster = max(cluster)

        else:
            raise ValueError('Clustering method %s not supported' %method)
    # return
    device = edge_index.device
    return torch.tensor(cluster).to(device)


def community_detection(edge_index,num_nodes,edge_attr=None,method='mcl'):
    """Detects clusters of nodes based on the edge attributes (distances)

    Args:
        edge_index (Tensor): Edge index
        num_nodes (int): Number of nodes
        edge_attr (Tensor, optional): Edge attributes. Defaults to None.
        method (str, optional): method. Defaults to 'mcl'.

    Raises:
        ValueError: Requires a valid clustering method ('mcl' or 'louvain')

    Returns:
        cluster Tensor
    """
    # make the networkX graph
    g = nx.Graph()
    g.add_nodes_from(range(num_nodes))

    for iedge,(i,j) in enumerate(edge_index.transpose(0,1).tolist()):
        if edge_attr is None:
            g.add_edge(i,j)
        else:
            g.add_edge(i,j,weight=edge_attr[iedge])

    # get the device
    device = edge_index.device


    # detect the communities using Louvain detection
    if method == 'louvain':
        cluster = community.best_partition(g)
        return torch.tensor([v for k,v in cluster.items()]).to(device)

    # detect the communities using MCL detection
    elif method == 'mcl':

        matrix = nx.to_scipy_sparse_matrix(g)

        result = mc.run_mcl(matrix)           # run MCL with default parameters

        clusters = mc.get_clusters(result)    # get clusters

        index = np.zeros(num_nodes).astype('int')
        for ic, c in enumerate(clusters):
            index[list(c)] = ic

        return torch.tensor(index).to(device)
    else:
        raise ValueError('Clustering method %s not supported' %method)

def community_pooling(cluster,data):
    """Pools features and edges of all cluster members 
    
    All cluster members are pooled into a single node that is assigned:
    - the max cluster value for each feature
    - the average cluster nodes position 

    Args:
        cluster ([type]): clusters
        data ([type]): features tensor

    Returns:
        pooled features tensor
    """
    # determine what the batches has as attributes
    has_internal_edges = hasattr(data,'internal_edge_index')
    has_pos2D = hasattr(data,'pos2D')
    has_pos = hasattr(data,'pos')
    has_cluster = hasattr(data,'cluster0')

    cluster, perm = consecutive_cluster(cluster)
    cluster = cluster.to(data.x.device)

    # pool the node infos
    x, _ = scatter_max(data.x,cluster,dim=0)

    # pool the edges
    edge_index, edge_attr  = pool_edge(cluster,data.edge_index, data.edge_attr)

    # pool internal edges if necessary
    if has_internal_edges:
        internal_edge_index, internal_edge_attr = pool_edge(cluster,data.internal_edge_index, data.internal_edge_attr)

    # pool the pos
    if has_pos:
        pos = scatter_mean(data.pos, cluster, dim=0)
    if has_pos2D:
        pos2D = scatter_mean(data.pos2D, cluster, dim=0)

    if has_cluster:
        c0, c1 = data.cluster0, data.cluster1


    # pool batch
    if hasattr(data,'batch'):
        batch = None if data.batch is None else pool_batch(perm, data.batch)
        data = Batch(batch=batch, x=x, edge_index=edge_index,
                     edge_attr = edge_attr, pos = pos)

        if has_internal_edges:
            data.internal_edge_index = internal_edge_index
            data.internal_edge_attr = internal_edge_attr

        if has_cluster:
            data.cluster0 = c0
            data.cluster1 = c1

    else:
        data = Data(x=x, edge_index=edge_index,
                     edge_attr = edge_attr, pos = pos)

        if has_internal_edges:
            data.internal_edge_index = internal_edge_index
            data.internal_edge_attr = internal_edge_attr

        if has_pos2D:
            data.pos2D = pos2D

        if has_cluster:
            data.cluster0 = c0
            data.cluster1 = c1

    return data


if __name__ == "__main__":

    import torch
    from torch_geometric.data import Data, Batch

    edge_index = torch.tensor([[0,1,1,2,3,4,4,5],
                               [1,0,2,1,4,3,5,4]],dtype=torch.long)

    x = torch.tensor([[0],[1],[2],[3],[4],[5]], dtype=torch.float)
    data = Data(x=x, edge_index=edge_index)
    data.pos = torch.tensor(np.random.rand(data.num_nodes,3))
    c = community_detection(data.edge_index,data.num_nodes)

    batch = Batch().from_data_list([data,data])
    cluster = community_detection(batch.edge_index,batch.num_nodes)
    new_batch = community_pooling(cluster,batch)


import torch
from torch.nn import Parameter
import torch.nn.functional as F
from torch_scatter import scatter_add
from torch_geometric.utils import remove_self_loops, add_self_loops, softmax

from torch_geometric.nn.inits import uniform
from torch_scatter import scatter_mean


def add_self_loops_wattr(edge_index, edge_attr, num_nodes=None):
    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    dtype, device = edge_index.dtype, edge_index.device
    loop = torch.arange(0, num_nodes, dtype=dtype, device=device)
    loop = loop.unsqueeze(0).repeat(2, 1)
    edge_index = torch.cat([edge_index, loop], dim=1)

    dtype, device = edge_attr.dtype, edge_attr.device
    loop = torch.ones(num_nodes,dtype=dtype, device=device)
    edge_attr = torch.cat([edge_attr,loop])

    return edge_index, edge_attr


class WGATConv(torch.nn.Module):

    """
    Args:
        in_channels (int): Size of each input sample.
        out_channels (int): Size of each output sample.
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 bias=True):

        super(WGATConv, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.weight = Parameter(
            torch.Tensor(2 * in_channels, out_channels))

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        size = 2 * self.in_channels
        uniform(size, self.weight)
        uniform(size, self.bias)

    def forward(self, x, edge_index, edge_attr):

        #print('weight : ', torch.sum(self.weight))

        row, col = edge_index
        num_node = len(x)
        edge_attr = edge_attr.unsqueeze(-1) if edge_attr.dim() == 1 else edge_attr

        # create edge feature by concatenating node feature
        alpha = torch.cat([x[row], x[col]], dim=-1)

        # multiply the edge features with the fliter
        alpha = torch.mm(alpha,self.weight)

        # multiply each edge features with the corresponding dist
        alpha = edge_attr*alpha

        # scatter the resulting edge feature to get node features
        out = torch.zeros(num_node,self.out_channels).to(alpha.device)
        out = scatter_mean(alpha,row,dim=0,out=out)
        out = scatter_mean(alpha,col,dim=0,out=out)

        # add the bias
        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                             self.in_channels,
                                             self.out_channels)



import unittest
from deeprank_gnn.GraphGenMP import GraphHDF5


class TestCreateGraph(unittest.TestCase):

    def setUp(self):

        self.pdb_path = './tests/data/pdb/1ATN/'
        self.pssm_path = './tests/data/pssm/1ATN/'
        self.ref = './tests/data/ref/1ATN/'

    def test_create(self):
        GraphHDF5(pdb_path=self.pdb_path, ref_path=self.ref, pssm_path=self.pssm_path,
                  graph_type='residue', outfile='1ATN_residue.hdf5',
                  nproc=1, tmpdir='./tmpdir')
